import abc

from gen_epix.casedb.domain import command, model
from gen_epix.casedb.domain.enum import ServiceType
from gen_epix.commondb.domain.command import Command
from gen_epix.commondb.services import AbacService as CommonAbacService


class BaseAbacService(CommonAbacService):
    SERVICE_TYPE = ServiceType.ABAC

    ORGANIZATION_ADMIN_WRITE_COMMANDS: set[type[Command]] = {  # type: ignore[assignment]
        command.COMMON_COMMAND_MAP.get(x, x)
        for x in CommonAbacService.ORGANIZATION_ADMIN_WRITE_COMMANDS
    } | {
        command.UserAccessCasePolicyCrudCommand,
        command.UserShareCasePolicyCrudCommand,
    }

    READ_ORGANIZATION_RESULTS_ONLY_COMMANDS: set[type[Command]] = {  # type: ignore[assignment]
        command.COMMON_COMMAND_MAP.get(x, x)
        for x in CommonAbacService.READ_ORGANIZATION_RESULTS_ONLY_COMMANDS
    } | {
        command.OrganizationAccessCasePolicyCrudCommand,
        command.OrganizationShareCasePolicyCrudCommand,
        command.UserAccessCasePolicyCrudCommand,
        command.UserShareCasePolicyCrudCommand,
    }

    READ_SELF_RESULTS_ONLY_COMMANDS: set[type[Command]] = {  # type: ignore[assignment]
        command.COMMON_COMMAND_MAP.get(x, x)
        for x in CommonAbacService.READ_SELF_RESULTS_ONLY_COMMANDS
    } | {
        command.UserAccessCasePolicyCrudCommand,
        command.UserShareCasePolicyCrudCommand,
    }

    READ_USER_COMMANDS: set[type[Command]] = {  # type: ignore[assignment]
        command.COMMON_COMMAND_MAP.get(x, x)
        for x in CommonAbacService.READ_USER_COMMANDS
    } | set()

    UPDATE_USER_COMMANDS: set[type[Command]] = {  # type: ignore[assignment]
        command.COMMON_COMMAND_MAP.get(x, x)
        for x in CommonAbacService.UPDATE_USER_COMMANDS
    } | set()

    CASE_ABAC_COMMANDS: set[type[Command]] = {
        command.RetrieveCompleteCaseTypeCommand,
        command.RetrieveCasesByQueryCommand,
        command.RetrieveCasesByIdCommand,
        command.RetrieveCaseRightsCommand,
        command.RetrieveCaseSetRightsCommand,
        command.RetrieveCaseTypeStatsCommand,
        command.RetrieveCaseSetStatsCommand,
        command.CaseTypeCrudCommand,
        command.CaseTypeSetMemberCrudCommand,
        command.CaseTypeSetCrudCommand,
        command.CaseTypeColCrudCommand,
        command.CaseTypeColSetCrudCommand,
        command.CaseTypeColSetMemberCrudCommand,
        command.CaseCrudCommand,
        command.CreateReadSetsForCasesCommand,
        command.CreateFileForReadSetCommand,
        command.CreateFileForSeqCommand,
        command.CreateSeqsForCasesCommand,
        # command.CaseDataCollectionUpdateAssociationCommand,
        command.CreateCaseSetCommand,
        command.CreateCasesCommand,
        command.CaseSetCrudCommand,
        # command.CaseSetCaseUpdateAssociationCommand,
        # command.CaseSetDataCollectionUpdateAssociationCommand,
        command.CaseDataCollectionLinkCrudCommand,
        command.CaseSetDataCollectionLinkCrudCommand,
        command.DataCollectionCrudCommand,
        command.RetrievePhylogeneticTreeByCasesCommand,
        command.RetrieveGeneticSequenceByCaseCommand,
        command.RetrieveGeneticSequenceFastaByCaseCommand,
        command.RetrieveCaseSetStatsCommand,
        command.RetrieveCaseTypeStatsCommand,
        command.ValidateCasesCommand,
    }

    @abc.abstractmethod
    def get_case_abac(self, cmd: command.Command) -> model.CaseAbac:
        raise NotImplementedError



from collections.abc import Callable
from typing import Annotated, Any, NoReturn, Self
from uuid import UUID

from fastapi import APIRouter, FastAPI, Form
from fastapi.responses import StreamingResponse
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field, model_validator

from gen_epix.casedb.domain import command, enum, model
from gen_epix.commondb.app_impl_details import AppImplDetails
from gen_epix.commondb.util import copy_model_field
from gen_epix.fastapp import App
from gen_epix.fastapp.api import CrudEndpointGenerator
from gen_epix.filter.datetime_range import TypedDatetimeRangeFilter


class UpdateCaseTypeSetCaseTypesRequestBody(PydanticBaseModel):
    case_type_set_members: list[model.CaseTypeSetMember]


class UpdateCaseTypeColSetCaseTypeColsRequestBody(PydanticBaseModel):
    case_type_col_set_members: list[model.CaseTypeColSetMember]


class ValidateCasesRequestBody(PydanticBaseModel):
    case_type_id: UUID = copy_model_field(command.ValidateCasesCommand, "case_type_id")
    created_in_data_collection_id: UUID = copy_model_field(
        command.ValidateCasesCommand, "created_in_data_collection_id"
    )
    data_collection_ids: set[UUID] = copy_model_field(
        command.ValidateCasesCommand, "data_collection_ids"
    )
    is_update: bool = copy_model_field(command.ValidateCasesCommand, "is_update")
    cases: list[model.CaseForCreateUpdate] = copy_model_field(
        command.ValidateCasesCommand, "cases"
    )

    @model_validator(mode="after")
    def _validate_cases(self) -> Self:
        if self.created_in_data_collection_id in self.data_collection_ids:
            raise ValueError(
                "The created in data collection ID may not be in the additional data collection IDs."
            )
        if self.is_update and any(x.id is None for x in self.cases):
            raise ValueError("All cases must have an ID when updating")
        return self


class CreateCasesRequestBody(PydanticBaseModel):
    case_type_id: UUID = copy_model_field(command.CreateCasesCommand, "case_type_id")
    created_in_data_collection_id: UUID = copy_model_field(
        command.CreateCasesCommand, "created_in_data_collection_id"
    )
    data_collection_ids: set[UUID] = copy_model_field(
        command.CreateCasesCommand, "data_collection_ids"
    )
    is_update: bool = copy_model_field(command.CreateCasesCommand, "is_update")
    cases: list[model.CaseForCreateUpdate] = copy_model_field(
        command.CreateCasesCommand, "cases"
    )

    @model_validator(mode="after")
    def _validate_cases(self) -> Self:
        if self.created_in_data_collection_id in self.data_collection_ids:
            raise ValueError(
                "The created in data collection ID may not be in the additional data collection IDs."
            )
        if self.is_update and any(x.id is None for x in self.cases):
            raise ValueError("All cases must have an ID when updating")
        return self


class CreateCaseSetRequestBody(PydanticBaseModel):
    case_set: model.CaseSet
    data_collection_ids: set[UUID] = Field(
        default=set(),
        description="The data collections in which the case set will be put initially",
    )
    case_ids: set[UUID] | None = Field(
        default=None, description="The cases to be added to the case set, if any."
    )


class RetrieveOrganizationContactRequestBody(PydanticBaseModel):
    organization_ids: list[UUID] | None = None
    site_ids: list[UUID] | None = None
    contact_ids: list[UUID] | None = None
    props: dict[str, Any] = {}

    @model_validator(mode="after")
    def check_one_of_fields(self) -> Any:
        if (
            sum(
                [
                    self.organization_ids is not None,
                    self.site_ids is not None,
                    self.contact_ids is not None,
                ]
            )
            != 1
        ):
            raise ValueError(
                "Exactly one of organization_ids, site_ids or contact_ids must be "
                "provided"
            )
        return self


class RetrievePhylogeneticTreeRequestBody(PydanticBaseModel):
    genetic_distance_case_type_col_id: UUID
    tree_algorithm_code: enum.TreeAlgorithmType
    case_ids: list[UUID]


class RetrieveGeneticSequenceRequestBody(PydanticBaseModel):
    genetic_sequence_case_type_col_id: UUID = Field(
        description="The case type column that contains the genetic sequences to retrieve.",
    )
    case_ids: list[UUID] = Field(
        description="The case ids to retrieve genetic sequences for.",
    )


class RetrieveAlleleProfileRequestBody(PydanticBaseModel):
    sequence_ids: list[UUID]
    props: dict[str, Any] = {}


class RetrieveCaseTypeStatsRequestBody(PydanticBaseModel):
    case_type_ids: set[UUID] | None = Field(
        default=None,
        description="The case type ids to retrieve stats for, if not all.",
    )
    datetime_range_filter: TypedDatetimeRangeFilter | None = Field(
        default=None,
        description="The datetime range to filter cases by, if any. The key attribute fo the filter should be left empty.",
    )


class RetrieveCaseSetStatsRequestBody(PydanticBaseModel):
    case_set_ids: set[UUID] | None = Field(
        default=None,
        description="The case set ids to retrieve stats for, if not all.",
    )


class CreateFileForForReadSetRequestBody(PydanticBaseModel):
    file_content: bytes = Field(description="The content of the file to create.")
    is_fwd: bool = Field(
        description="Whether the file is for the forward reads (True) or reverse reads (False).",
    )


class CreateFileForSeqRequestBody(PydanticBaseModel):
    file_content: bytes = Field(description="The content of the file to create.")


def create_case_endpoints(
    router: APIRouter | FastAPI,
    app: App,
    handle_exception: Callable[[str, Any, Exception], NoReturn] | None = None,
    **kwargs: Any,
) -> None:
    assert handle_exception
    app_impl: AppImplDetails = app.impl
    registered_user_dependency = app_impl.registered_user_dependency

    # Specific endpoints - Case
    @router.put(
        "/case_type_sets/{case_type_set_id}/case_types",
        operation_id="case_type_sets__put__case_types",
        name="Update association between CaseTypeSet and CaseType",
        description=command.CaseTypeSetCaseTypeUpdateAssociationCommand.__doc__,
    )
    async def case_type_sets__put__case_types(
        user: registered_user_dependency,  # type: ignore
        case_type_set_id: UUID,
        request_body: UpdateCaseTypeSetCaseTypesRequestBody,
    ) -> list[model.CaseSetMember]:
        try:
            cmd = command.CaseTypeSetCaseTypeUpdateAssociationCommand(
                user=user,
                obj_id1=case_type_set_id,
                association_objs=request_body.case_type_set_members,
                props={"return_id": False},
            )
            retval: list[model.CaseSetMember] = app.handle(cmd)
        except Exception as exception:
            handle_exception("fbe272b9", user, exception)
        return retval

    @router.put(
        "/case_type_col_sets/{case_type_col_set_id}/case_type_cols",
        operation_id="case_type_col_sets__put__case_type_cols",
        name="Update association between CaseTypeColSet and CaseTypeCol",
        description=command.CaseTypeColSetCaseTypeColUpdateAssociationCommand.__doc__,
    )
    async def case_type_col_sets__put__case_type_cols(
        user: registered_user_dependency,  # type: ignore
        case_type_col_set_id: UUID,
        request_body: UpdateCaseTypeColSetCaseTypeColsRequestBody,
    ) -> list[model.CaseTypeColSetMember]:
        try:
            cmd = command.CaseTypeColSetCaseTypeColUpdateAssociationCommand(
                user=user,
                obj_id1=case_type_col_set_id,
                association_objs=request_body.case_type_col_set_members,
                props={"return_id": False},
            )
            retval: list[model.CaseTypeColSetMember] = app.handle(cmd)
        except Exception as exception:
            handle_exception("ab010768", user, exception)
        return retval

    @router.get(
        "/complete_case_types",
        operation_id="complete_case_types__get_one",
        name="Retrieve complete case type",
        description=command.RetrieveCompleteCaseTypeCommand.__doc__,
    )
    async def complete_case_types__get_one(
        user: registered_user_dependency,  # type: ignore
        case_type_id: UUID,
    ) -> model.CompleteCaseType:
        try:
            cmd = command.RetrieveCompleteCaseTypeCommand(
                user=user, case_type_id=case_type_id
            )
            retval: model.CompleteCaseType = app.handle(cmd)
        except Exception as exception:
            handle_exception("c6c17125", user, exception)
        return retval

    @router.post(
        "/validate/cases",
        operation_id="validate__cases",
        name="Validate cases",
        description=command.ValidateCasesCommand.__doc__,
    )
    async def validate__cases(
        user: registered_user_dependency,  # type: ignore
        request_body: ValidateCasesRequestBody,
    ) -> model.CaseValidationReport:
        try:
            cmd = command.ValidateCasesCommand(
                user=user,
                case_type_id=request_body.case_type_id,
                created_in_data_collection_id=request_body.created_in_data_collection_id,
                is_update=request_body.is_update,
                cases=request_body.cases,
                data_collection_ids=request_body.data_collection_ids,
            )
            retval: model.CaseValidationReport = app.handle(cmd)
        except Exception as exception:
            handle_exception("9f8e7d6c", user, exception)
        return retval

    @router.post(
        "/create/cases",
        operation_id="create__cases",
        name="Create cases",
        description=command.CreateCasesCommand.__doc__,
    )
    async def create__cases(
        user: registered_user_dependency,  # type: ignore
        request_body: CreateCasesRequestBody,
    ) -> list[model.Case]:
        try:
            cmd = command.CreateCasesCommand(
                user=user,
                cases=request_body.cases,
                data_collection_ids=request_body.data_collection_ids,
                case_type_id=request_body.case_type_id,
                created_in_data_collection_id=request_body.created_in_data_collection_id,
                is_update=request_body.is_update,
            )
            retval: list[model.Case] = app.handle(cmd)
        except Exception as exception:
            handle_exception("b413ab76", user, exception)
        return retval

    @router.post(
        "/create/case_set",
        operation_id="create__case_set",
        name="Create case set",
        description=command.CreateCaseSetCommand.__doc__,
    )
    async def create__case_set(
        user: registered_user_dependency,  # type: ignore
        request_body: CreateCaseSetRequestBody,
    ) -> model.CaseSet:
        try:
            cmd = command.CreateCaseSetCommand(
                user=user,
                case_set=request_body.case_set,
                data_collection_ids=request_body.data_collection_ids,
                case_ids=request_body.case_ids,
            )
            retval: model.CaseSet = app.handle(cmd)
        except Exception as exception:
            handle_exception("c39c42f9", user, exception)
        return retval

    @router.post(
        "/retrieve/case_type_stats",
        operation_id="retrieve__case_type_stats",
        name="Retrieve case type statistics",
        description=command.RetrieveCaseTypeStatsCommand.__doc__,
    )
    async def retrieve__case_type_stats(
        user: registered_user_dependency,  # type: ignore
        request_body: RetrieveCaseTypeStatsRequestBody,
    ) -> list[model.CaseTypeStat]:
        try:
            cmd = command.RetrieveCaseTypeStatsCommand(
                user=user,
                case_type_ids=request_body.case_type_ids,
                datetime_range_filter=request_body.datetime_range_filter,
            )
            retval: list[model.CaseTypeStat] = app.handle(cmd)
        except Exception as exception:
            handle_exception("80c99f53", user, exception)
        return retval

    @router.post(
        "/retrieve/case_set_stats",
        operation_id="retrieve__case_set_stats",
        name="Retrieve case set statistics",
        description=command.RetrieveCaseSetStatsCommand.__doc__,
    )
    async def retrieve__case_set_stats(
        user: registered_user_dependency,  # type: ignore
        request_body: RetrieveCaseSetStatsRequestBody,
    ) -> list[model.CaseSetStat]:
        try:
            cmd = command.RetrieveCaseSetStatsCommand(
                user=user,
                case_set_ids=(
                    None
                    if not request_body.case_set_ids
                    else list(request_body.case_set_ids)
                ),
            )
            retval: list[model.CaseSetStat] = app.handle(cmd)
        except Exception as exception:
            handle_exception("be54843e", user, exception)
        return retval

    @router.post(
        "/retrieve/case_ids_by_query",
        operation_id="retrieve__case_ids_by_query",
        name="Retrieve case IDs by query",
        description=command.RetrieveCasesByQueryCommand.__doc__,
    )
    async def retrieve__case_ids_by_query(
        user: registered_user_dependency,  # type: ignore
        request_body: model.CaseQuery,
    ) -> list[UUID]:
        try:
            retval: list[UUID] = app.handle(
                command.RetrieveCasesByQueryCommand(
                    user=user,
                    case_query=request_body,
                )
            )
        except Exception as exception:
            handle_exception("a8f773fe", user, exception)
        return retval

    @router.post(
        "/retrieve/cases_by_ids",
        operation_id="retrieve__cases_by_ids",
        name="Retrieve cases by IDs",
        description=command.RetrieveCasesByIdCommand.__doc__,
    )
    async def retrieve__cases_by_ids(
        user: registered_user_dependency,  # type: ignore
        request_body: list[UUID],
    ) -> list[model.Case]:
        try:
            retval: list[model.Case] = app.handle(
                command.RetrieveCasesByIdCommand(
                    user=user,
                    case_ids=request_body,
                )
            )
        except Exception as exception:
            handle_exception("f6d423fe", user, exception)
        return retval

    @router.post(
        "/retrieve/case_rights",
        operation_id="retrieve__case_rights",
        name="Retrieve case rights",
        description=command.RetrieveCaseRightsCommand.__doc__,
    )
    async def retrieve__case_rights(
        user: registered_user_dependency,  # type: ignore
        request_body: list[UUID],
    ) -> list[model.CaseRights]:
        try:
            retval: list[model.CaseRights] = app.handle(
                command.RetrieveCaseRightsCommand(
                    user=user,
                    case_ids=request_body,
                )
            )
        except Exception as exception:
            handle_exception("c6f4b3c2", user, exception)
        return retval

    @router.post(
        "/retrieve/case_set_rights",
        operation_id="retrieve__case_set_rights",
        name="Retrieve case set rights",
        description=command.RetrieveCaseSetRightsCommand.__doc__,
    )
    async def retrieve__case_set_rights(
        user: registered_user_dependency,  # type: ignore
        request_body: list[UUID],
    ) -> list[model.CaseSetRights]:
        try:
            retval: list[model.CaseSetRights] = app.handle(
                command.RetrieveCaseSetRightsCommand(
                    user=user,
                    case_set_ids=request_body,
                )
            )
        except Exception as exception:
            handle_exception("b9c49fe1", user, exception)
        return retval

    @router.post(
        "/retrieve/organization_contact",
        operation_id="retrieve__organization_contact",
        name="Retrieve organization contact",
        description=command.RetrieveOrganizationContactCommand.__doc__,
    )
    async def retrieve__organization_contact(
        user: registered_user_dependency,  # type: ignore
        request_body: RetrieveOrganizationContactRequestBody,
    ) -> list[model.Contact]:
        try:
            retval: list[model.Contact] = app.handle(
                command.RetrieveOrganizationContactCommand(
                    user=user,
                    organization_ids=request_body.organization_ids,
                    site_ids=request_body.site_ids,
                    contact_ids=request_body.contact_ids,
                    props=request_body.props,
                )
            )
        except Exception as exception:
            handle_exception(  # type:ignore[call-arg]
                "b8172f62",
                user,
                exception,
                request_ids=(request_body.organization_ids or [])
                + (request_body.site_ids or [])
                + (request_body.contact_ids or []),
            )
        return retval

    @router.post(
        "/retrieve/phylogenetic_tree",
        operation_id="retrieve__phylogenetic_tree",
        name="Retrieve phylogenetic tree",
        description=command.RetrievePhylogeneticTreeByCasesCommand.__doc__,
    )
    async def retrieve__phylogenetic_tree(
        user: registered_user_dependency, request_body: RetrievePhylogeneticTreeRequestBody  # type: ignore
    ) -> model.PhylogeneticTree:
        try:
            retval: model.PhylogeneticTree = app.handle(
                command.RetrievePhylogeneticTreeByCasesCommand(
                    user=user,
                    genetic_distance_case_type_col_id=request_body.genetic_distance_case_type_col_id,
                    tree_algorithm=request_body.tree_algorithm_code,
                    case_ids=request_body.case_ids,
                )
            )
        except Exception as exception:
            handle_exception(  # type:ignore[call-arg]
                "45219a88", user, exception, request_ids=request_body.case_ids
            )
        return retval

    @router.post(
        "/retrieve/genetic_sequence",
        operation_id="retrieve__genetic_sequence",
        name="Retrieve genetic sequence by case",
        description=command.RetrieveGeneticSequenceByCaseCommand.__doc__,
    )
    async def retrieve__genetic_sequence(
        user: registered_user_dependency,  # type: ignore
        request_body: RetrieveGeneticSequenceRequestBody,
    ) -> list[model.GeneticSequence]:
        try:
            retval: list[model.GeneticSequence] = app.handle(
                command.RetrieveGeneticSequenceByCaseCommand(
                    user=user,
                    genetic_sequence_case_type_col_id=request_body.genetic_sequence_case_type_col_id,
                    case_ids=request_body.case_ids,
                )
            )
        except Exception as exception:
            handle_exception(  # type:ignore[call-arg]
                "1238afb2", user, exception, request_ids=request_body.case_ids
            )
        return retval

    @router.post(
        "/retrieve/genetic_sequence/fasta",
        operation_id="retrieve__genetic_sequence__fasta",
        name="Retrieve genetic sequence by case, in fasta format and streamed",
        description=command.RetrieveGeneticSequenceFastaByCaseCommand.__doc__,
    )
    async def retrieve__genetic_sequence_fasta(
        token: Annotated[str, Form()],
        genetic_sequence_case_type_col_id: Annotated[str, Form()],
        case_ids: Annotated[list[UUID], Form()],
        file_name: Annotated[str, Form()],
    ) -> StreamingResponse:
        user: model.User | None = None
        app_impl: AppImplDetails = app.impl
        try:
            user = await app_impl.services[
                enum.ServiceType.AUTH
            ].get_existing_user_from_token(token=token)
            fasta_iterable = app.handle(
                command.RetrieveGeneticSequenceFastaByCaseCommand(
                    user=user,
                    genetic_sequence_case_type_col_id=(
                        genetic_sequence_case_type_col_id
                    ),
                    case_ids=case_ids,
                )
            )
        except Exception as exception:
            handle_exception(  # type:ignore[call-arg]
                "d4c2e1b1",
                user,
                exception,
                request_ids=case_ids,
            )

        return StreamingResponse(
            fasta_iterable,
            media_type="application/x-fasta",
            headers={"Content-Disposition": f'attachment; filename="{file_name}"'},
        )

    @router.post(
        "/retrieve/allele_profile",
        operation_id="retrieve__allele_profile",
        name="Retrieve allele profile",
        description=command.RetrieveAlleleProfileCommand.__doc__,
    )
    async def retrieve__allele_profile(
        user: registered_user_dependency, request_body: RetrieveAlleleProfileRequestBody  # type: ignore
    ) -> list[model.AlleleProfile]:
        try:
            retval: list[model.AlleleProfile] = app.handle(
                command.RetrieveAlleleProfileCommand(
                    user=user,
                    sequence_ids=request_body.sequence_ids,
                    props=request_body.props,
                )
            )
        except Exception as exception:
            handle_exception(  # type:ignore[call-arg]
                "a4c03b54", user, exception, request_ids=request_body.sequence_ids
            )
        return retval

    @router.post(
        "/create_read_sets_for_cases",
        operation_id="create__read_sets_for_cases",
        name="Create reads sets for cases",
        description=command.CreateReadSetsForCasesCommand.__doc__,
    )
    async def create__read_sets_for_cases(
        user: registered_user_dependency,  # type: ignore
        case_read_sets: list[model.CaseReadSet],
    ) -> list[model.ReadSet]:
        try:
            created_read_sets: list[model.ReadSet] = app.handle(
                command.CreateReadSetsForCasesCommand(
                    user=user,
                    case_read_sets=case_read_sets,
                )
            )
        except Exception as exception:
            handle_exception("e3d4f5a6", user, exception)
        return created_read_sets

    @router.post(
        "/create_file_for_read_set/{case_id}/{case_type_col_id}",
        operation_id="create_file_for_read_set",
        name="Create file for reads set",
        description=command.CreateFileForReadSetCommand.__doc__,
    )
    async def create_file_for_read_set(
        user: registered_user_dependency,  # type: ignore
        case_id: UUID,
        case_type_col_id: UUID,
        request_body: CreateFileForForReadSetRequestBody,
    ) -> UUID:
        try:
            created_file_id: UUID = app.handle(
                command.CreateFileForReadSetCommand(
                    user=user,
                    file_content=request_body.file_content,
                    case_id=case_id,
                    case_type_col_id=case_type_col_id,
                    is_fwd=request_body.is_fwd,
                )
            )
        except Exception as exception:
            handle_exception("d3f4e2b1", user, exception)
        return created_file_id

    @router.post(
        "/create_seqs_for_cases",
        operation_id="create_seqs_for_cases",
        name="Create sequences for cases",
        description=command.CreateSeqsForCasesCommand.__doc__,
    )
    async def create_seqs_for_cases(
        user: registered_user_dependency,  # type: ignore
        case_seqs: list[model.CaseSeq],
    ) -> list[model.Seq]:
        try:
            created_seqs: list[model.Seq] = app.handle(
                command.CreateSeqsForCasesCommand(
                    user=user,
                    case_seqs=case_seqs,
                )
            )
        except Exception as exception:
            handle_exception("a1b2c3d4", user, exception)
        return created_seqs

    @router.post(
        "/create_file_for_seq/{case_id}/{case_type_col_id}",
        operation_id="create_file_for_seq",
        name="Create file for sequence",
        description=command.CreateFileForSeqCommand.__doc__,
    )
    async def create_file_for_seq(
        user: registered_user_dependency,  # type: ignore
        case_id: UUID,
        case_type_col_id: UUID,
        request_body: CreateFileForSeqRequestBody,
    ) -> UUID:
        try:
            created_file_id: UUID = app.handle(
                command.CreateFileForSeqCommand(
                    user=user,
                    file_content=request_body.file_content,
                    case_id=case_id,
                    case_type_col_id=case_type_col_id,
                )
            )
        except Exception as exception:
            handle_exception("b5c6d7e8", user, exception)
        return created_file_id

    @router.get(
        "/retrieve/library_prep_protocols",
        operation_id="retrieve__library_prep_protocols",
        name="Retrieve library preparation protocols",
        description=command.RetrieveLibraryPrepProtocolsCommand.__doc__,
    )
    async def retrieve__library_prep_protocols(
        user: registered_user_dependency,  # type: ignore
    ) -> list[model.LibraryPrepProtocol]:
        try:
            retval: list[model.LibraryPrepProtocol] = app.handle(
                command.RetrieveLibraryPrepProtocolsCommand(
                    user=user,
                )
            )
        except Exception as exception:
            handle_exception("e7f8a9b0", user, exception)
        return retval

    @router.get(
        "/retrieve/assembly_protocols",
        operation_id="retrieve__assembly_protocols",
        name="Retrieve assembly protocols",
        description=command.RetrieveAssemblyProtocolsCommand.__doc__,
    )
    async def retrieve__assembly_protocols(
        user: registered_user_dependency,  # type: ignore
    ) -> list[model.AssemblyProtocol]:
        try:
            retval: list[model.AssemblyProtocol] = app.handle(
                command.RetrieveAssemblyProtocolsCommand(
                    user=user,
                )
            )
        except Exception as exception:
            handle_exception("c1d2e3f4", user, exception)
        return retval

    # CRUD
    crud_endpoint_sets = CrudEndpointGenerator.create_crud_endpoint_set_for_domain(
        app,
        service_type=enum.ServiceType.CASE,
        user_dependency=registered_user_dependency,
    )
    CrudEndpointGenerator.generate_endpoints(
        router, crud_endpoint_sets, handle_exception
    )



from collections.abc import Iterable
from uuid import UUID

import gen_epix.seqdb.domain.command as seqdb_command
from gen_epix.casedb.domain import command, enum, exc, model
from gen_epix.casedb.domain.policy.abac import BaseCaseAbacPolicy
from gen_epix.casedb.services.case.base import BaseCaseService
from gen_epix.fastapp.enum import CrudOperation
from gen_epix.fastapp.unit_of_work import BaseUnitOfWork


def case_service_retrieve_phylogenetic_tree(
    self: BaseCaseService, cmd: command.RetrievePhylogeneticTreeByCasesCommand
) -> model.PhylogeneticTree:
    dist_case_type_col_id = cmd.genetic_distance_case_type_col_id
    tree_algorithm_code = cmd.tree_algorithm
    case_ids = cmd.case_ids
    user: model.User
    user, repository = self._get_user_and_repository(cmd)  # type: ignore[assignment]
    assert isinstance(user, model.User) and user.id is not None
    case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)
    assert case_abac is not None

    with repository.uow() as uow:
        # Get distance column data
        dist_case_type_col: model.CaseTypeCol = repository.crud(  # type: ignore[assignment]
            uow,
            user.id,
            model.CaseTypeCol,
            None,
            dist_case_type_col_id,
            CrudOperation.READ_ONE,
        )
        case_type_id = dist_case_type_col.case_type_id
        dist_col: model.Col = repository.crud(  # type: ignore[assignment]
            uow,
            user.id,
            model.Col,
            None,
            dist_case_type_col.col_id,
            CrudOperation.READ_ONE,
        )
        if dist_col.col_type != enum.ColType.GENETIC_DISTANCE:
            raise exc.InvalidArgumentsError(
                f"Case type column {dist_case_type_col_id} is not of type {enum.ColType.GENETIC_DISTANCE.value}"
            )
        # Get sequence column data
        seq_case_type_col_id = dist_case_type_col.genetic_sequence_case_type_col_id
        if not seq_case_type_col_id:
            raise exc.InvalidArgumentsError(
                f"Case type column {dist_case_type_col_id} has no associated sequence column"
            )

        # @ABAC
        assert dist_case_type_col.tree_algorithm_codes is not None
        if tree_algorithm_code not in dist_case_type_col.tree_algorithm_codes:
            raise exc.UnauthorizedAuthError(
                f"User {user.id} has no read access to tree algorithm {tree_algorithm_code}"
            )

        # Get genetic distance protocol
        genetic_distance_protocol: model.GeneticDistanceProtocol = (
            self.repository.crud(  # type: ignore[assignment]
                uow,
                user.id,
                model.GeneticDistanceProtocol,
                None,
                dist_col.genetic_distance_protocol_id,
                CrudOperation.READ_ONE,
            )
        )
        seqdb_seq_distance_protocol_id = (
            genetic_distance_protocol.seqdb_seq_distance_protocol_id
        )

        # Special case: zero case_ids
        if not case_ids:
            retval: model.PhylogeneticTree = self.app.handle(
                command.RetrievePhylogeneticTreeBySequencesCommand(
                    user=user,
                    tree_algorithm_code=tree_algorithm_code,
                    seqdb_seq_distance_protocol_id=seqdb_seq_distance_protocol_id,
                    sequence_ids=[],
                )
            )
            retval.genetic_distance_protocol_id = genetic_distance_protocol.id
            return retval

        # Create temporary case_abac only for this case type and the
        # seq_case_type_col_id having the same rights as the dist_case_type_col
        temp_case_abac = model.CaseAbac(
            is_full_access=case_abac.is_full_access,
            case_type_access_abacs={},
            case_type_share_abacs={},
        )
        for data_collection_id, x in case_abac.case_type_access_abacs.get(
            case_type_id, {}
        ).items():
            if dist_case_type_col_id not in x.read_case_type_col_ids:
                continue
            if case_type_id not in temp_case_abac.case_type_access_abacs:
                temp_case_abac.case_type_access_abacs[case_type_id] = {}
            temp_case_abac.case_type_access_abacs[case_type_id][data_collection_id] = (
                model.CaseTypeAccessAbac(
                    read_case_type_col_ids={seq_case_type_col_id},
                    **x.model_dump(exclude={"read_case_type_col_ids"}),
                )
            )

        # @ABAC: Get cases
        cases = self._retrieve_cases_with_content_right(
            uow,
            user.id,
            temp_case_abac,
            enum.CaseRight.READ_CASE,
            case_ids=case_ids,
            case_type_ids={case_type_id},
            filter_content=True,
        )

        # Get sequence_ids from seq_case_type_col
        case_sequence_map = {}
        for case in cases:
            sequence_id = case.content.get(seq_case_type_col_id)
            if sequence_id:
                case_sequence_map[case.id] = UUID(sequence_id)

        # Retrieve tree and remove sequence_ids to avoid leaking information
        sequence_ids = list(case_sequence_map.values())
        sequence_case_map = {y: x for x, y in case_sequence_map.items()}
        phylogenetic_tree: model.PhylogeneticTree = self.app.handle(
            command.RetrievePhylogeneticTreeBySequencesCommand(
                user=cmd.user,
                tree_algorithm_code=tree_algorithm_code,
                seqdb_seq_distance_protocol_id=seqdb_seq_distance_protocol_id,
                sequence_ids=sequence_ids,
                props={
                    "leaf_id_mapper": lambda x: sequence_case_map[x],
                },
            )
        )
        phylogenetic_tree.genetic_distance_protocol_id = genetic_distance_protocol.id
        phylogenetic_tree.sequence_ids = None

    return phylogenetic_tree


def case_service_retrieve_genetic_sequence_by_case(
    self: BaseCaseService,
    cmd: command.RetrieveGeneticSequenceByCaseCommand,
) -> list[model.GeneticSequence]:
    seq_case_type_col_id = cmd.genetic_sequence_case_type_col_id
    case_ids = cmd.case_ids
    user, repository = self._get_user_and_repository(cmd)
    assert isinstance(user, model.User) and user.id is not None
    if not case_ids:
        return []

    case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)
    assert case_abac is not None

    with repository.uow() as uow:
        seq_ids_or_none: list[UUID | None] = _get_seq_ids_from_cases(
            self,
            uow,
            user,
            case_abac,
            case_ids,
            seq_case_type_col_id,
        )
        if any(x is None for x in seq_ids_or_none):
            raise exc.NoResultsError("Not all cases have a sequence")
        seq_ids: list[UUID] = seq_ids_or_none  # type: ignore[assignment]

        # Retrieve sequences
        genetic_sequences: list[model.GeneticSequence] = self.app.handle(
            command.RetrieveGeneticSequenceByIdCommand(
                user=user,
                seq_ids=seq_ids,
            )
        )

    return genetic_sequences


def case_service_retrieve_genetic_sequence_fasta_by_case(
    self: BaseCaseService, cmd: command.RetrieveGeneticSequenceFastaByCaseCommand
) -> Iterable[str]:
    """
    Return a streaming iterable of FASTA formatted lines.
    Path:
    HTTP client
    -> casedb endpoint
    -> casedb service calls casedb seqdb command
    -> seqdb command (inside casedb) calls ext_app with RetrieveSeqFastaCommand
    -> seqdb service calls correct repository (dict or SA implementation) to stream Seq rows
    -> seqdb service converts rows to FASTA lines on the fly
    -> returns an iterator
    -> casedb forwards that iterator
    -> FastAPI wraps it in a StreamingResponse.
    """
    seq_case_type_col_id = cmd.genetic_sequence_case_type_col_id
    case_ids = cmd.case_ids
    user, repository = self._get_user_and_repository(cmd)
    assert isinstance(user, model.User) and user.id is not None

    if not case_ids:
        raise exc.InvalidArgumentsError("No case ids given")

    case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)
    assert case_abac is not None

    with repository.uow() as uow:

        seq_ids_or_none: list[UUID | None] = _get_seq_ids_from_cases(
            self,
            uow,
            user,
            case_abac,
            case_ids,
            seq_case_type_col_id,
        )
        if any(x is None for x in seq_ids_or_none):
            raise exc.NoResultsError("Not all cases have a sequence")
        seq_ids: list[UUID] = seq_ids_or_none  # type: ignore[assignment]
        retrieve_cmd = command.RetrieveGeneticSequenceFastaByIdCommand(
            user=cmd.user, seq_ids=seq_ids
        )
        fasta_iterator: Iterable[str] = self.app.handle(retrieve_cmd)
        return fasta_iterator


def case_service_retrieve_library_prep_protocols(
    self: BaseCaseService,
    cmd: command.RetrieveLibraryPrepProtocolsCommand,
) -> list[model.LibraryPrepProtocol]:
    user, repository = self._get_user_and_repository(cmd)
    assert isinstance(user, model.User) and user.id is not None

    library_prep_protocols: list[model.LibraryPrepProtocol] = self.app.handle(
        seqdb_command.LibraryPrepProtocolCrudCommand(
            user=cmd.user,
            operation=CrudOperation.READ_ALL,
        )
    )
    return library_prep_protocols


def case_service_retrieve_assembly_protocols(
    self: BaseCaseService, cmd: command.RetrieveAssemblyProtocolsCommand
) -> list[model.AssemblyProtocol]:
    user, repository = self._get_user_and_repository(cmd)
    assert isinstance(user, model.User) and user.id is not None

    assembly_protocols: list[model.AssemblyProtocol] = self.app.handle(
        seqdb_command.AssemblyProtocolCrudCommand(
            user=cmd.user,
            operation=CrudOperation.READ_ALL,
        )
    )
    return assembly_protocols


def _get_seq_ids_from_cases(
    self: BaseCaseService,
    uow: BaseUnitOfWork,
    user: model.User,
    case_abac: model.CaseAbac,
    case_ids: list[UUID],
    seq_case_type_col_id: UUID,
) -> list[UUID | None]:
    # @ABAC: Get cases and sequence_ids
    cases = self._retrieve_cases_with_content_right(
        uow,
        user.id,  # type: ignore[arg-type]
        case_abac,
        enum.CaseRight.READ_CASE,
        case_ids=case_ids,
        filter_content=True,
    )
    retval = [x.content.get(seq_case_type_col_id) for x in cases]
    return [UUID(x) for x in retval if x]



from collections.abc import Hashable, Iterable
from typing import Any

from gen_epix.commondb.domain import model
from gen_epix.commondb.domain.model import Model
from gen_epix.commondb.domain.repository.organization import BaseOrganizationRepository
from gen_epix.fastapp import Entity, exc
from gen_epix.fastapp.repositories import DictRepository
from gen_epix.fastapp.unit_of_work import BaseUnitOfWork


class OrganizationDictRepository(DictRepository, BaseOrganizationRepository):
    def __init__(
        self,
        entities: Iterable[Entity],
        db: dict[type[Model], dict[Hashable, Model]],
        user_class: type[model.User] = model.User,
        user_invitation_class: type[model.UserInvitation] = model.UserInvitation,
        **kwargs: Any,
    ):
        BaseOrganizationRepository.__init__(
            self, user_class=user_class, user_invitation_class=user_invitation_class
        )
        DictRepository.__init__(self, entities, db, **kwargs)

    def is_existing_user_by_key(
        self, uow: BaseUnitOfWork, user_key: str | None
    ) -> bool:
        if user_key is None:
            return False
        for user in self._db[self.user_class].values():
            assert isinstance(user, self.user_class)
            if user.key == user_key.lower():
                return True
        return False

    def retrieve_user_by_key(self, uow: BaseUnitOfWork, user_key: str) -> model.User:
        for user in self._db[self.user_class].values():
            assert isinstance(user, self.user_class)
            if user.key == user_key.lower():
                return user
        raise exc.NoResultsError()



import datetime
from typing import Literal

from pydantic import Field

from gen_epix.filter.enum import FilterType
from gen_epix.filter.range import RangeFilter


class DatetimeRangeFilter(RangeFilter):
    lower_bound: datetime.datetime | None = Field(
        default=None, description="The lower bound of the range.", frozen=True
    )
    upper_bound: datetime.datetime | None = Field(
        default=None, description="The upper bound of the range.", frozen=True
    )


class TypedDatetimeRangeFilter(DatetimeRangeFilter):
    type: Literal[FilterType.DATETIME_RANGE.value]



"""
Object adapters for providing unified interface across different object types.
"""

from collections.abc import Hashable, Iterator
from typing import Any, Protocol, runtime_checkable

from pydantic import BaseModel


@runtime_checkable
class RowLike(Protocol):
    """Protocol for row-like objects."""

    def get(self, key: Hashable, default: Any = None) -> Any:
        """Get value by key with optional default."""
        ...

    def __getitem__(self, key: Hashable) -> Any:
        """Get value by key."""
        ...

    def __setitem__(self, key: Hashable, value: Any) -> None:
        """Set value by key."""
        ...

    def __contains__(self, key: Hashable) -> bool:
        """Check if key exists."""
        ...

    def keys(self) -> Iterator[Hashable]:
        """Get all keys."""
        ...


class DictAdapter:
    """Adapter for dictionary objects."""

    def __init__(self, obj: dict):
        self._obj = obj

    def get(self, key: Hashable, default: Any = None) -> Any:
        """Get value by key with optional default."""
        return self._obj.get(key, default)

    def set(self, key: Hashable, value: Any) -> None:
        """Set value by key."""
        self._obj[key] = value

    def has_key(self, key: Hashable) -> bool:
        """Check if key exists."""
        return key in self._obj

    def keys(self) -> Iterator[Hashable]:
        """Get all keys."""
        return iter(self._obj.keys())


class PydanticAdapter:
    """Adapter for Pydantic model objects."""

    def __init__(self, obj: BaseModel):
        self._obj = obj

    def get(self, key: Hashable, default: Any = None) -> Any:
        """Get value by key with optional default."""
        return getattr(self._obj, str(key), default)

    def set(self, key: Hashable, value: Any) -> None:
        """Set value by key."""
        setattr(self._obj, str(key), value)

    def has_key(self, key: Hashable) -> bool:
        """Check if key exists."""
        return hasattr(self._obj, str(key))

    def keys(self) -> Iterator[Hashable]:
        """Get all keys."""
        return iter(self._obj.model_fields.keys())


class PolarsAdapter:
    """Adapter for Polars objects."""

    def __init__(self, obj: Any):
        self._obj = obj

    def get(self, key: Hashable, default: Any = None) -> Any:
        """Get value by key with optional default."""
        try:
            return self._obj[key]
        except (KeyError, IndexError):
            return default

    def set(self, key: Hashable, value: Any) -> None:
        """Set value by key."""
        self._obj = self._obj.with_columns({str(key): value})

    def has_key(self, key: Hashable) -> bool:
        """Check if key exists."""
        return str(key) in self._obj.columns

    def keys(self) -> Iterator[Hashable]:
        """Get all keys."""
        return iter(self._obj.columns)


class ObjectAdapter:
    """
    Unified adapter that provides consistent interface for different object types.

    Supports dict, Pydantic models, and Polars objects.
    """

    def __init__(self, obj: dict | BaseModel | Any):
        self._obj = obj
        self._adapter = self._create_adapter(obj)

    def _create_adapter(
        self, obj: Any
    ) -> DictAdapter | PydanticAdapter | PolarsAdapter:
        """Factory method to create appropriate adapter for object type."""
        if isinstance(obj, dict):
            return DictAdapter(obj)
        elif isinstance(obj, BaseModel):
            return PydanticAdapter(obj)
        elif hasattr(obj, "__dataframe__") or hasattr(
            obj, "columns"
        ):  # Polars detection
            return PolarsAdapter(obj)
        else:
            raise ValueError(f"Unsupported object type: {type(obj)}")

    def get(self, key: Hashable, default: Any = None) -> Any:
        """Get value by key with optional default."""
        return self._adapter.get(key, default)

    def set(self, key: Hashable, value: Any) -> None:
        """Set value by key."""
        self._adapter.set(key, value)

    def has_key(self, key: Hashable) -> bool:
        """Check if key exists."""
        return self._adapter.has_key(key)

    def keys(self) -> Iterator[Hashable]:
        """Get all keys."""
        return self._adapter.keys()

    def unwrap(self) -> Any:
        """Return the original object."""
        return self._obj



import re
import uuid
from collections.abc import Callable, Hashable, Mapping
from enum import Enum
from functools import partial
from typing import Any, ClassVar, Self
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator

from gen_epix.fastapp.domain.key import Key
from gen_epix.fastapp.domain.link import Link
from gen_epix.fastapp.enum import FieldType, StringCasing
from gen_epix.fastapp.exc import DomainException


class Entity(BaseModel):
    CAMEL_TO_SNAKE_CASE_PATTERN: ClassVar[re.Pattern] = re.compile(
        r"(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])"
    )
    NO_MODEL_ERROR_MSG: ClassVar[str] = "Entity does not have a model set"
    NOT_PERSISTABLE_ERROR_MSG: ClassVar[str] = "Entity is not persistable"
    NO_ERM_ERROR_MSG: ClassVar[str] = (
        "Entity is not linked to an entity-relationship model"
    )
    DEFAULT_ID_FIELD_NAME: ClassVar[str] = "id"

    model_config = ConfigDict(arbitrary_types_allowed=True)

    id: UUID = Field(default_factory=uuid.uuid4)
    persistable: bool = Field(default=False, frozen=True)
    snake_case_singular_name: str | None = None
    snake_case_plural_name: str | None = None
    camel_case_singular_name: str | None = None
    camel_case_plural_name: str | None = None
    pascal_case_singular_name: str | None = None
    pascal_case_plural_name: str | None = None
    kebab_case_singular_name: str | None = None
    kebab_case_plural_name: str | None = None
    url_name: str | None = None
    database_name: str | None = None
    schema_name: str | None = None
    table_name: str | None = None
    id_field_name: str | None = None
    keys: dict[int, Key] = {}
    links: dict[int, Link] = {}

    _model_class: type[BaseModel] | None = None
    _db_model_class: type | None = None
    _crud_command_class: type | None = None
    _create_api_model_class: type | None = None
    _read_api_model_class: type | None = None
    _fields: dict[str, dict[str, Any]] | None = None
    _links_by_field_name: dict[str, tuple[int, Link]] = {}
    _get_link_id_by_model_class: dict[type, Callable[[Any], Hashable]] = {}
    _keys_generator: Callable | None = None
    _has_model: bool = False

    @field_validator(
        "snake_case_plural_name",
        "snake_case_singular_name",
        "camel_case_plural_name",
        "camel_case_singular_name",
        "pascal_case_plural_name",
        "pascal_case_singular_name",
        "kebab_case_plural_name",
        "kebab_case_singular_name",
        "url_name",
        "database_name",
        "schema_name",
        "table_name",
        "id_field_name",
        mode="before",
    )
    @classmethod
    def _validate_names(cls, value: str | Enum | None) -> str | None:
        return str(value.value) if isinstance(value, Enum) else value

    @model_validator(mode="before")
    @classmethod
    def _validate_model(cls, data: Any) -> Any:
        if (
            "id_field_name" not in data
            and "persistable" in data
            and data["persistable"]
        ):
            data["id_field_name"] = cls.DEFAULT_ID_FIELD_NAME
        return data

    @field_validator("keys", mode="before")
    @classmethod
    def _validate_keys(cls, value: dict[int, Any]) -> dict[int, Key]:
        """
        Validate and convert keys to Key objs.
        """
        return {x: y if isinstance(y, Key) else Key(y) for x, y in value.items()}

    @field_validator("links", mode="before")
    @classmethod
    def _validate_links(cls, value: dict[int, Any]) -> dict[int, Link]:
        """
        Validate and convert links to Link objs.
        """
        return {
            x: (
                y
                if isinstance(y, Link)
                else (
                    Link(
                        link_field_name=y[0],
                        link_model_class=y[1],
                        relationship_field_name=y[2],
                    )
                    if isinstance(y, tuple)
                    else Link(**y)
                )
            )
            for x, y in value.items()
        }

    @property
    def name(self) -> str:
        if not self.has_model():
            raise DomainException(Entity.NO_MODEL_ERROR_MSG)
        return self._model_class.NAME  # type: ignore

    @property
    def model_class(self) -> type[BaseModel]:
        if not self.has_model():
            raise DomainException(Entity.NO_MODEL_ERROR_MSG)
        return self._model_class  # type: ignore

    @property
    def crud_command_class(self) -> type[BaseModel] | None:
        if not self.has_model():
            raise DomainException(Entity.NO_MODEL_ERROR_MSG)
        if not self.persistable:
            raise ValueError(Entity.NOT_PERSISTABLE_ERROR_MSG)
        return self._crud_command_class  # type: ignore

    @property
    def db_model_class(self) -> type | None:
        if not self.has_model():
            raise DomainException(Entity.NO_MODEL_ERROR_MSG)
        if not self.persistable:
            raise ValueError(Entity.NOT_PERSISTABLE_ERROR_MSG)
        return self._db_model_class  # type: ignore

    @property
    def create_api_model_class(self) -> type | None:
        if not self.has_model():
            raise DomainException(Entity.NO_MODEL_ERROR_MSG)
        return self._create_api_model_class  # type: ignore

    @property
    def read_api_model_class(self) -> type | None:
        if not self.has_model():
            raise DomainException(Entity.NO_MODEL_ERROR_MSG)
        return self._read_api_model_class  # type: ignore

    @property
    def get_obj_id(self) -> Callable[[Any], Hashable]:
        if not self.has_model():
            raise DomainException(Entity.NO_MODEL_ERROR_MSG)
        assert self.id_field_name
        return lambda x: getattr(x, self.id_field_name)

    def set_model_class(
        self, model_class: type[BaseModel], on_existing: str = "raise"
    ) -> Self:
        """
        Set the model class for the entity.

        Parameters
        ----------
        model_class : type[BaseModel]
            The model class to set for the entity.

        Returns
        -------
        Self
            The updated entity obj.

        Raises
        ------
        ValueError
            If the entity already has a model set or if the id_field_name is invalid.
        """
        if self.has_model():
            if self._model_class is model_class:
                return self
            if on_existing == "raise":
                raise ValueError("Instance already has a model set")
            elif on_existing == "replace":
                pass
            else:
                raise ValueError(f"Unknown on_existing value: {on_existing}")
        field_names = Entity._get_model_field_names(model_class)

        # Set fields and derive their FieldType
        self._fields = {
            x[0]: {
                "name": x[0],
                "alias": x[0] if not x[1] else x[1],
                "type": FieldType.VALUE,
            }
            for x in field_names
        }

        # Set ID field
        if self.id_field_name:
            if self.id_field_name not in self._fields:
                raise ValueError(
                    f"id_field_name property for model {model_class} does "
                    f"not contain a valid field name: {self.id_field_name}"
                )
            self._fields[self.id_field_name]["type"] = FieldType.ID

        # Set LINK and RELATIONSHIP fields
        for link in self.links.values():
            self._fields[link.link_field_name]["type"] = FieldType.LINK
            if link.relationship_field_name:
                self._fields[link.relationship_field_name][
                    "type"
                ] = FieldType.RELATIONSHIP

        # Set COMPUTED fields
        for field_name in model_class.model_computed_fields:
            self._fields[field_name]["type"] = FieldType.COMPUTED

        # Verify and parse links
        self._verify_and_parse_model_links(model_class)

        # Set properties and create keys generator
        self._model_class = model_class
        self._has_model = True
        self._create_keys_generator()
        return self

    def has_model(self) -> bool:
        """
        Check if the entity has a model set.

        Returns
        -------
        bool
            True if the entity has a model set, False otherwise.
        """
        return self._model_class is not None

    def set_db_model_class(self, db_model_class: type) -> Self:
        """
        Set the repository model class for the entity, which is intended as the class
        that is stored or retrieved from the repository.
        """
        if not self.has_model():
            raise DomainException(Entity.NO_MODEL_ERROR_MSG)
        if not self.persistable:
            raise ValueError(Entity.NOT_PERSISTABLE_ERROR_MSG)
        self._db_model_class = db_model_class
        return self

    def set_create_api_model_class(self, create_api_model_class: type) -> Self:
        """
        Set the API model class for the entity, which is intended as the request model
        that is posted to an endpoint to create a resource.
        """
        if not self.has_model():
            raise DomainException(Entity.NO_MODEL_ERROR_MSG)
        self._create_api_model_class = create_api_model_class
        return self

    def set_read_api_model_class(self, read_api_model_class: type) -> Self:
        """
        Set the API model class for the entity, which is intended as the model
        that returned in a response by an endpoint.
        """
        if not self.has_model():
            raise DomainException(Entity.NO_MODEL_ERROR_MSG)
        self._read_api_model_class = read_api_model_class
        return self

    def set_crud_command_class(self, crud_command_class: type[BaseModel]) -> Self:
        """
        Set the CRUD command class for the entity, which is intended as the class
        that is stored or retrieved from the repository.
        """
        if not self.has_model():
            raise DomainException(Entity.NO_MODEL_ERROR_MSG)
        if not self.persistable:
            raise ValueError(Entity.NOT_PERSISTABLE_ERROR_MSG)
        self._crud_command_class = crud_command_class
        return self

    def get_field_names(
        self, by_alias: bool = True, field_type: FieldType | None = None
    ) -> list[str]:
        """
        Get the field names of the entity.

        Parameters
        ----------
        by_alias : bool, optional
            If True, return the field names by alias. Defaults to True.
        field_type : FieldType or None, optional
            The type of model fields to filter by. Defaults to None.

        Returns
        -------
        list[str]
            A list of field names of the entity.

        Raises
        ------
        ValueError
            If the entity does not have fields set.
        """
        if self._fields is None:
            raise ValueError("Entity does not have fields set")
        key = "alias" if by_alias else "name"
        return [
            x[key]
            for x in self._fields.values()
            if not field_type or x["type"] == field_type
        ]

    def get_id_field_name(self, by_alias: bool = True) -> str | None:
        """
        Get the ID field name of the entity.

        Parameters
        ----------
        by_alias : bool, optional
            If True, return the ID field name by alias. Defaults to True.

        Returns
        -------
        str
            The ID field name of the entity.

        """
        field_names = self.get_field_names(by_alias=by_alias, field_type=FieldType.ID)
        if not field_names:
            raise ValueError("Entity does not have an ID field")
        return field_names[0]

    def get_keys_field_names(self, by_alias: bool = True) -> list[tuple[str, ...]]:
        if by_alias:
            return [
                tuple(self._fields[y]["alias"] for y in x.field_names)
                for x in self.keys.values()
            ]
        return [x.field_names for x in self.keys.values()]

    def get_link_field_names(self, by_alias: bool = True) -> list[str]:
        """
        Get the link field names of the entity.

        Parameters
        ----------
        by_alias : bool, optional
            If True, return the link field names by alias. Defaults to True.

        Returns
        -------
        list[str]
            A list of link field names of the entity.
        """

        return self.get_field_names(by_alias=by_alias, field_type=FieldType.LINK)

    def get_relationship_field_names(self, by_alias: bool = True) -> list[str]:
        """
        Get the relationship field names of the entity.

        Parameters
        ----------
        by_alias : bool, optional
            If True, return the back-populate field names by alias. Defaults to True.

        Returns
        -------
        list[str]
            A list of back-populate field names of the entity.
        """
        return self.get_field_names(
            by_alias=by_alias, field_type=FieldType.RELATIONSHIP
        )

    def get_value_field_names(self, by_alias: bool = True) -> list[str]:
        """
        Get the value field names of the entity.

        Parameters
        ----------
        by_alias : bool, optional
            If True, return the value field names by alias. Defaults to True.

        Returns
        -------
        list[str]
            A list of value field names of the entity.
        """

        return self.get_field_names(by_alias=by_alias, field_type=FieldType.VALUE)

    def get_keys_generator(self) -> Callable | None:
        """
        Get the keys generator for the entity.

        Returns
        -------
        Callable or None
            The keys generator for the entity, or None if no keys generator is set.

        Raises
        ------
        ValueError
            If the entity does not have a model set.
        """
        if not self.has_model():
            raise ValueError("Entity does not have a model set")
        return self._keys_generator

    def get_link_id(self, link_model_class: type) -> Callable[[Any], Hashable]:
        fun = self._get_link_id_by_model_class.get(link_model_class)
        if fun is None:
            raise ValueError(
                f"No link or several links to {link_model_class.__name__} exist for entity {self.name}"
            )
        return fun

    def get_link_entity(self, link_field_name: str) -> Any | None:
        """
        Get the linked entity, if any.

        Parameters
        ----------
        link_field_name : str
            The field name of the link.

        Returns
        -------
        Entity | None
            The linked entity or None if no linked entity.

        Raises
        ------
        ValueError
            If the entity is not linked to an entity-relationship model.
        """
        _, link = self._links_by_field_name.get(link_field_name, (None, None))
        return link.link_model_class.ENTITY if link else None  # type: ignore

    def get_link_properties_by_field_name(
        self, link_field_name: str
    ) -> tuple[int, type[BaseModel], str | None]:
        """
        Get the properties of a link by its field name.

        Parameters
        ----------
        link_field_name : str
            The field name of the link.

        Returns
        -------
        tuple[int, type[BaseModel], str or None]
            A tuple containing the link type ID, link model class, and back-populate
            field name.

        Raises
        ------
        ValueError
            If the entity is not linked to an entity-relationship model or if the
            entity does not have links set.
        """
        link_type_id, link = self._links_by_field_name.get(
            link_field_name, (None, None)
        )
        if not link_type_id or not link:
            raise ValueError(f"Field {link_field_name} is not a link field")
        return (link_type_id, link.link_model_class, link.relationship_field_name)

    def has_links(self) -> bool:
        """
        Check if the entity has links.

        Returns
        -------
        bool
            True if the entity has links, False otherwise.
        """
        return len(self.links) > 0

    def has_keys(self) -> bool:
        """
        Check if the entity has keys.

        Returns
        -------
        bool
            True if the entity has keys, False otherwise.
        """
        return len(self.keys) > 0

    def get_name_by_casing(
        self,
        string_casing: StringCasing = StringCasing.SNAKE_CASE,
        is_plural: bool = False,
    ) -> str | None:
        if string_casing == StringCasing.SNAKE_CASE:
            name = (
                self.snake_case_plural_name
                if is_plural
                else self.snake_case_singular_name
            )
        elif string_casing == StringCasing.PASCAL_CASE:
            name = (
                self.pascal_case_plural_name
                if is_plural
                else self.pascal_case_singular_name
            )
        elif string_casing == StringCasing.CAMEL_CASE:
            name = (
                self.camel_case_plural_name
                if is_plural
                else self.camel_case_singular_name
            )
        else:
            raise NotImplementedError(f"String casing {string_casing} not implemented")
        return name

    def copy(self, update: Mapping[str, Any]) -> "Entity":
        """
        Create a copy of the entity with the given keyword arguments replacing
        any of the existing values.

        Parameters
        ----------
        **kwargs : Any
            Keyword arguments to update the entity.

        Returns
        -------
        Self
            A copy of the entity with updated attributes.
        """
        props = self.model_dump(exclude_unset=True, exclude_defaults=True)
        props.update(update)
        return Entity(**props)

    def _verify_and_parse_model_links(self, model_class: type[BaseModel]) -> Self:
        """
        Check if the link field names and back populate field names are valid given
        the model class. If they are valid, set the type of the field to LINK and
        add a map from link field name to link type id and link object to the entity.

        Parameters
        ----------
        model_class : type[BaseModel]
            The model class to verify and parse links for.

        Returns
        -------
        Self
            The updated entity obj.

        Raises
        ------
        ValueError
            If any link field names or back populate field names are invalid.
        """
        link_field_names: set[str] = set()
        relationship_field_names: set[str] = set()
        to_remove_get_link_id = set()
        assert isinstance(self._fields, dict)
        assert isinstance(self._links_by_field_name, dict)
        assert isinstance(self._get_link_id_by_model_class, dict)
        for link_type_id, link in self.links.items():
            self._verify_link_field_name(model_class, link_field_names, link)
            self._fields[link.link_field_name]["type"] = FieldType.LINK
            self._links_by_field_name[link.link_field_name] = (link_type_id, link)
            if link.link_model_class in self._get_link_id_by_model_class:
                to_remove_get_link_id.add(link.link_model_class)
            else:
                self._get_link_id_by_model_class[link.link_model_class] = partial(
                    lambda x, y: getattr(y, x), link.link_field_name
                )
            self._verify_relationship_fields(
                model_class, relationship_field_names, link
            )
        for link_model_class in to_remove_get_link_id:
            self._get_link_id_by_model_class.pop(link_model_class)
        self._check_identical_field_names(link_field_names, relationship_field_names)
        return self

    def _verify_link_field_name(
        self, model_class: type[BaseModel], link_field_names: set, link: Link
    ) -> None:
        """
        Verify the link field name.

        Parameters
        ----------
        model_class : type[BaseModel]
            The model class to verify the link field name for.
        link_field_names : set
            The set of link field names.
        link : Link
            The link to verify.

        Raises
        ------
        ValueError
            If the link field name is invalid or not unique.
        """
        assert self._fields
        if link.link_field_name not in self._fields:
            raise ValueError(
                f"Link field name {link.link_field_name} to {link.link_model_class} "
                f"for model {model_class.__name__} is not a valid field name"
            )
        if link.link_field_name == self.id_field_name:
            raise ValueError(
                f"Link field name is identical to id field name: {link.link_field_name}"
            )
        if link.link_field_name in link_field_names:
            raise ValueError(
                f"Link field name {link.link_field_name}"
                f" for model {model_class.__name__} is not unique"
            )
        link_field_names.add(link.link_field_name)

    def _verify_relationship_fields(
        self, model_class: type[BaseModel], relationship_field_names: set, link: Link
    ) -> None:
        """
        Verify the back populate field names.

        Parameters
        ----------
        model_class : type[BaseModel]
            The model class to verify the back populate field names for.
        relationship_field_names : set
            The set of back populate field names.
        link : Link
            The link to verify.

        Raises
        ------
        ValueError
            If the back populate field name is invalid or not unique.
        """
        if link.relationship_field_name:
            assert self._fields
            if link.relationship_field_name not in self._fields:
                raise ValueError(
                    f"Back populate field name {link.relationship_field_name} "
                    f"to {link.link_field_name} for model {model_class.__name__} "
                    "is not a valid field name"
                )
            if link.relationship_field_name == self.id_field_name:
                raise ValueError(
                    "Back populate field name is identical to id "
                    f"field name: {link.relationship_field_name}"
                )
            if link.relationship_field_name in relationship_field_names:
                raise ValueError(
                    f"Back populate field name {link.relationship_field_name}"
                    f" for model {model_class.__name__} is not unique"
                )
            relationship_field_names.add(link.relationship_field_name)
            self._fields[link.relationship_field_name]["type"] = FieldType.RELATIONSHIP

    def _check_identical_field_names(
        self, link_field_names: set, relationship_field_names: set
    ) -> None:
        identical_field_names = link_field_names & relationship_field_names
        if identical_field_names:
            identical_field_names_str = ", ".join(identical_field_names)
            raise ValueError(
                "Some back populate field names "
                f"are identical to link field names: {identical_field_names_str}"
            )

    def _create_keys_generator(self) -> None:
        """Create a function that will calculate all the keys for a given obj"""
        # Special case: no keys
        if not self.keys:
            self._keys_generator = lambda x: {}
            return

        def keys_generator(keys: dict[int, Key], obj: BaseModel) -> dict[int, Any]:
            return {x: y(obj) for x, y in keys.items()}

        self._keys_generator = partial(keys_generator, self.keys)

    @classmethod
    def camel_to_snake_case(cls, value: str) -> str:
        return cls.CAMEL_TO_SNAKE_CASE_PATTERN.sub("_", value).lower()

    @staticmethod
    def _get_model_field_names(
        model_class: type[BaseModel],
    ) -> list[tuple[str, str | None]]:
        """
        Returns a list of (field_name, alias) tuples.
        """
        field_names = []
        for field_name, field in (
            model_class.model_fields | model_class.model_computed_fields
        ).items():
            field_names.append((field_name, field.alias))
        return field_names

    def __hash__(self) -> int:
        return self.name.__hash__()



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later

import datetime
from collections.abc import Callable
from typing import Any
from uuid import UUID

from pydantic import Field, field_serializer

from gen_epix.commondb.domain import model
from gen_epix.commondb.util import generate_ulid
from gen_epix.fastapp import Command as ServiceCommand
from gen_epix.fastapp import CrudCommand as ServiceCrudCommand
from gen_epix.fastapp import UpdateAssociationCommand as ServiceUpdateAssociationCommand


class Command(ServiceCommand):
    id: UUID = Field(default_factory=generate_ulid, description="The ID of the command")
    created_at: datetime.datetime = Field(
        default_factory=datetime.datetime.now,
        description="The created timestamp of the command",
    )
    user: model.User | None = None
    props: dict[str, Any] = {}

    @field_serializer("created_at", mode="plain")
    def _serialize_created_at(self, value: datetime.datetime) -> str | None:
        return value.isoformat() if value else None

    @field_serializer("props", mode="plain")
    def _serialize_props(self, value: dict[str, Any]) -> dict[str, Any]:
        return {x: y for x, y in value.items() if not isinstance(y, Callable)}


class CrudCommand(ServiceCrudCommand, Command):
    user: model.User | None = None
    obj_ids: UUID | list[UUID] | None = None  # type: ignore


class UpdateAssociationCommand(ServiceUpdateAssociationCommand, Command):
    user: model.User | None = None
    obj_id1: UUID | list[UUID] | None = None
    obj_id2: UUID | list[UUID] | None = None
    association_objs: list[model.Model] | None = None



from types import TracebackType
from typing import Self

from pydantic import ValidationError as PydanticValidationError
from sqlalchemy import exc as sa_exc
from sqlalchemy.orm import Session

from gen_epix.fastapp import exc
from gen_epix.fastapp.unit_of_work import BaseUnitOfWork


class SAUnitOfWork(BaseUnitOfWork):
    """
    Unit of work class wrapping the SQLAlchemy session.

    The context stack that can be passed during construction indicates whether work
    would be executed within another unit of work's context. If so, that context will
    be responsible for committing or rolling back the session and this unit of work will
    not commit or rollback on exit when used as a context manager. This avoids creating
    nested sessions.
    """

    def __init__(
        self, session: Session, context_stack: list[BaseUnitOfWork] | None = None
    ):
        super().__init__()
        self._session = session
        self._context_stack = context_stack

    @property
    def session(self) -> Session:
        return self._session

    def commit(self) -> None:
        self._session.commit()

    def rollback(self) -> None:
        self._session.rollback()

    def flush(self) -> None:
        self._session.flush()

    @staticmethod
    def _handle_exception(
        exception_class: type[Exception],
        exception_value: Exception,
        traceback: TracebackType | None,
    ) -> Exception:
        """
        Handle exceptions raised during a unit of work, converting them into a domain
        exception when they are due to normal operation such as a unique constraint
        failing. Any other exceptions are re-raised with traceback.
        """
        if issubclass(exception_class, exc.DomainException):
            raise exception_value
        if issubclass(exception_class, AttributeError):
            raise exception_value
        if issubclass(exception_class, PydanticValidationError):
            raise ValueError(
                f"Model validation error: {str(exception_value)}"
            ).with_traceback(traceback)
        if issubclass(exception_class, sa_exc.IntegrityError):
            if "UNIQUE" in str(exception_value).upper():
                raise exc.UniqueConstraintViolationError(
                    f"Unique constraint violation: {str(exception_value)}",
                )
            elif "NOT NULL" in str(exception_value).upper():
                raise exc.NotNullConstraintViolationError(
                    f"Not null constraint violation: {str(exception_value)}",
                )
            elif "FOREIGN KEY" in str(exception_value).upper():
                raise exc.LinkConstraintViolationError(
                    f"Foreign key constraint violation: {str(exception_value)}",
                )
            else:
                raise NotImplementedError().with_traceback(traceback)
        if issubclass(exception_class, sa_exc.StatementError):
            raise ValueError(f"Statement error: {str(exception_value)}").with_traceback(
                traceback
            )
        raise NotImplementedError().with_traceback(traceback)

    def __enter__(self) -> Self:
        if self._context_stack is not None:
            self._context_stack.append(self)
        self._is_managing_context = True
        return self

    def __exit__(
        self,
        exception_class: type[Exception] | None,
        exception_value: Exception | None,
        traceback: TracebackType | None,
    ) -> None:
        self._is_managing_context = False
        # Handle nested contexts
        if self._context_stack is not None:
            # Remove self from stack
            assert self._context_stack[-1] is self
            self._context_stack.pop()
            # Check if nested context
            if self._context_stack:
                # Nested context since stack is not empty -> do not commit or rollback,
                # let the outer context handle it instead
                return
        # Commit or rollback based on exception
        if exception_class is None:
            try:
                self.commit()
            except Exception as exception:
                self.rollback()
                # Propagate exception
                SAUnitOfWork._handle_exception(
                    type(exception), exception, exception.__traceback__
                )
        else:
            self.rollback()
            # Propagate exception
            SAUnitOfWork._handle_exception(exception_class, exception_value, traceback)  # type: ignore[arg-type]



import logging
from typing import Any

from gen_epix.commondb.domain.model import User
from gen_epix.fastapp import App


def set_log_level(
    prefix: str, log_level: int = logging.ERROR, delimiter: str = "."
) -> None:
    def _set_log_level(logger_name: str) -> None:
        logger = logging.getLogger(logger_name)
        logger.setLevel(log_level)
        for handler in logger.handlers:
            handler.setLevel(log_level)

    for suffix in ["setup", "app", "service"]:
        _set_log_level(f"{prefix}{delimiter}{suffix}")
    # Other loggers
    _set_log_level("httpx")


def create_root_user_from_claims(cfg: dict, app: App) -> User:
    user_manager = app.user_manager
    if user_manager is None:
        raise ValueError("User generator not found")
    root_user_cfg: dict = cfg["service"]["auth"]["props"]["root"]["user"]
    user: User = user_manager.create_root_user_from_claims(  # type:ignore[assignment]
        {
            "__key__": root_user_cfg["key"],
            "email": root_user_cfg.get("email"),
            "name": root_user_cfg.get("name"),
        },
    )
    if user.email and not user.name:
        user.name = user.email.split("@")[0]

    return user


def parse_stats(df: list[dict], stats: Any, **kwargs: Any) -> None:
    for (
        function_name,
        function_profile,
    ) in stats.get_stats_profile().func_profiles.items():
        row = {**kwargs}
        n_calls = function_profile.ncalls.split("/")
        if len(n_calls) == 1:
            n_calls_min = n_calls[0]
            n_calls_max = n_calls[0]
        else:
            n_calls_min = min(n_calls)
            n_calls_max = max(n_calls)
        row.update(
            {
                "function_name": function_name,
                "n_calls_min": n_calls_min,
                "n_calls_max": n_calls_max,
                "total_time": function_profile.tottime,
                "total_time_per_call": function_profile.percall_tottime,
                "cumulative_time": function_profile.cumtime,
                "cumulative_time_per_call": function_profile.cumtime,
                "file_name": function_profile.file_name,
                "line_number": function_profile.line_number,
            }
        )
        df.append(row)



from uuid import UUID

from gen_epix.casedb.domain import command, model
from gen_epix.casedb.domain.policy import BaseCaseAbacPolicy
from gen_epix.casedb.services.case.base import BaseCaseService
from gen_epix.fastapp.enum import CrudOperation
from gen_epix.filter import UuidSetFilter


def case_service_retrieve_complete_case_type(
    self: BaseCaseService,
    cmd: command.RetrieveCompleteCaseTypeCommand,
) -> model.CompleteCaseType:
    # TODO: many calls are inefficient,
    # retrieving first all objs and then filtering.
    # To be improved with e.g. CQS.
    user, repository = self._get_user_and_repository(cmd)

    with repository.uow() as uow:
        # Get case type
        case_type_id = cmd.case_type_id
        case_type: model.CaseType = self.repository.crud(  # type: ignore[assignment]
            uow,
            user.id,
            model.CaseType,
            None,
            case_type_id,
            CrudOperation.READ_ONE,
        )

        # @ABAC
        # Get allowed case type columns with any CRUD permission
        case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)
        assert case_abac is not None
        case_type_access_abacs: dict[UUID, model.CaseTypeAccessAbac] = (
            case_abac.case_type_access_abacs.get(case_type_id, {})
        )
        case_type_share_abacs: dict[UUID, model.CaseTypeShareAbac] = (
            case_abac.case_type_share_abacs.get(case_type_id, {})
        )

        abac_case_type_col_ids: set[UUID]
        if case_abac.is_full_access:
            # Special case: full access -> all rights for all data collections for
            # this case type
            # TODO: consider if it should be limited to the union of all the
            # organization rights instead. A root user e.g. may then still have
            # full access by using the CRUD methods
            abac_case_type_col_ids = repository.crud(  # type: ignore[assignment]
                uow,
                user.id,
                model.CaseTypeCol,
                None,
                None,
                CrudOperation.READ_ALL,
                filter=UuidSetFilter(
                    key="case_type_id",
                    members=frozenset({case_type_id}),
                ),
                return_id=True,
            )
            data_collection_ids: list[UUID] = self.app.handle(
                command.DataCollectionCrudCommand(
                    user=user,  # type: ignore[arg-type]
                    operation=CrudOperation.READ_ALL,
                    props={"return_id": True},
                )
            )
            case_type_access_abacs = {
                x: model.CaseTypeAccessAbac(
                    case_type_id=case_type_id,
                    data_collection_id=x,
                    is_private=True,
                    add_case=True,
                    remove_case=True,
                    read_case_type_col_ids=abac_case_type_col_ids,
                    write_case_type_col_ids=abac_case_type_col_ids,
                    add_case_set=True,
                    remove_case_set=True,
                    read_case_set=True,
                    write_case_set=True,
                )
                for x in data_collection_ids
            }
            # case_type_share_abacs can be empty since all rights are already in
            # case_type_access_abacs
            case_type_share_abacs = {}
        else:
            abac_case_type_col_ids = set()
            for x in case_type_access_abacs.values():
                abac_case_type_col_ids.update(x.read_case_type_col_ids)
                abac_case_type_col_ids.update(x.write_case_type_col_ids)

        # Get etiologies
        if case_type.disease_id:
            etiologies = self.app.handle(
                command.EtiologyCrudCommand(
                    user=user,  # type: ignore[arg-type]
                    operation=CrudOperation.READ_ALL,
                )
            )
            etiologies = {
                x.id: x for x in etiologies if x.disease_id == case_type.disease_id
            }
        else:
            etiologies = {}

        # Get etiological agents
        if etiologies:
            etiological_agent_ids = list(
                x.etiological_agent_id for x in etiologies.values()
            )
            etiological_agents = self.app.handle(
                command.EtiologicalAgentCrudCommand(
                    user=user,  # type: ignore[arg-type]
                    operation=CrudOperation.READ_SOME,
                    obj_ids=etiological_agent_ids,
                )
            )
            etiological_agents = {x.id: x for x in etiological_agents}
        else:
            etiological_agents = {}

        # Get allowed case_type_cols
        case_type_col_ids = list(abac_case_type_col_ids)
        case_type_cols_: list[model.CaseTypeCol] = repository.crud(  # type: ignore[assignment]
            uow,
            user.id,
            model.CaseTypeCol,
            None,
            case_type_col_ids,
            CrudOperation.READ_SOME,
        )
        case_type_cols: dict[UUID, model.CaseTypeCol] = {
            x.id: x for x in case_type_cols_  # type: ignore[misc]
        }

        # # Special case: no case_type_cols
        # if not case_type_cols:
        #     return model.CompleteCaseType(
        #         **case_type.model_dump(),
        #         etiologies=etiologies,
        #         etiological_agents=etiological_agents,
        #         dims={},
        #         cols={},
        #         case_type_dims=[],
        #         case_type_cols={},
        #         case_type_col_order=[],
        #         genetic_distance_protocols={},
        #         tree_algorithms={},
        #         case_type_access_abacs=case_type_access_abacs,
        #     )

        # Get cols
        col_ids = list({x.col_id for x in case_type_cols.values()})
        cols_: list[model.Col] = repository.crud(  # type: ignore[assignment]
            uow,
            user.id,
            model.Col,
            None,
            col_ids,
            CrudOperation.READ_SOME,
        )
        cols: dict[UUID, model.Col] = {x.id: x for x in cols_}  # type: ignore[misc]

        # Get dims
        dim_ids = list({x.dim_id for x in cols.values()})
        dims_: list[model.Dim] = repository.crud(  # type: ignore[assignment]
            uow,
            user.id,
            model.Dim,
            None,
            dim_ids,
            CrudOperation.READ_SOME,
        )
        dims: dict[UUID, model.Dim] = {x.id: x for x in dims_}  # type: ignore[misc]

        # Get case_type_col_order
        # TODO: to be tested
        max_dim_rank = max([0] + [x.rank for x in dims.values() if x.rank])
        max_col_rank_in_dim = max(
            [0] + [x.rank_in_dim for x in cols.values() if x.rank_in_dim]
        )
        max_case_type_col_rank = max(
            [0] + [x.rank for x in case_type_cols.values() if x.rank]
        )
        max_case_type_col_occurrence = max(
            [0] + [x.occurrence for x in case_type_cols.values() if x.occurrence]
        )
        case_type_col_keys: dict[UUID, tuple[int, int, int]] = {
            x.id: (  # type: ignore[misc]
                x.rank if x.rank else max_case_type_col_rank,
                (
                    dims[cols[x.col_id].dim_id].rank
                    if dims[cols[x.col_id].dim_id].rank
                    else max_dim_rank
                ),
                (
                    cols[x.col_id].rank_in_dim
                    if cols[x.col_id].rank_in_dim
                    else max_col_rank_in_dim
                ),
                x.occurrence if x.occurrence else max_case_type_col_occurrence,
            )
            for x in case_type_cols.values()
        }
        case_type_col_order = list(case_type_col_keys.keys())
        case_type_col_order.sort(key=lambda x: case_type_col_keys[x])

        # Get case_type_dims as the list ordered by the (dim, occurrence)
        # that occurs first in case_type_col_order
        dict_: dict[tuple[UUID, int | None], list] = {}
        # dict[tuple[dim_id, occurrence], list[rank, [tuple[case_type_col_id, col.rank_in_dim]]]]
        for case_type_col_id in case_type_col_order:
            # Add to dict_
            case_type_col = case_type_cols[case_type_col_id]
            col = cols[case_type_col.col_id]
            tuple_ = (col.dim_id, case_type_col.occurrence)
            if tuple_ in dict_:
                dict_[tuple_][1].append((case_type_col_id, col.rank_in_dim))
                continue
            dict_[tuple_] = [len(dict_), [(case_type_col_id, col.rank_in_dim)]]
        case_type_dim_order = list(dict_.keys())
        case_type_dim_order.sort(key=lambda x: dict_[x][0])
        case_type_dims = [
            model.CaseTypeDim(
                id=x[0],
                dim_id=x[0],
                occurrence=x[1],
                rank=i + 1,
                case_type_col_order=[],
            )
            for i, x in enumerate(case_type_dim_order)
        ]
        for case_type_dim in case_type_dims:
            # Fill in id and case_type_col_order
            tuples = dict_[(case_type_dim.dim_id, case_type_dim.occurrence)][1]
            tuples.sort(key=lambda x: 1 if x[1] is None else x[1])
            case_type_dim.case_type_col_order = [x[0] for x in tuples]
            case_type_dim.id = case_type_dim.case_type_col_order[0]

        # Get genetic distance protocols
        genetic_distance_protocols = self.app.handle(
            command.GeneticDistanceProtocolCrudCommand(
                user=user,  # type: ignore[arg-type]
                operation=CrudOperation.READ_SOME,
                obj_ids=list(
                    {
                        x.genetic_distance_protocol_id
                        for x in cols.values()
                        if x.genetic_distance_protocol_id
                    }
                ),
            )
        )
        genetic_distance_protocols = {x.id: x for x in genetic_distance_protocols}

        # Get tree algorithms
        tree_algorithm_codes = set.union(
            set(),
            *[
                x.tree_algorithm_codes
                for x in case_type_cols.values()
                if x.tree_algorithm_codes
            ],
        )
        tree_algorithms = self.app.handle(
            command.TreeAlgorithmCrudCommand(
                user=user,  # type: ignore[arg-type]
                operation=CrudOperation.READ_ALL,
            )
        )
        tree_algorithms = {
            x.code: x for x in tree_algorithms if x.code in tree_algorithm_codes
        }

    # Compose complete case type and return
    return model.CompleteCaseType(
        **case_type.model_dump(),
        etiologies=etiologies,
        etiological_agents=etiological_agents,
        dims=dims,
        cols=cols,
        case_type_dims=case_type_dims,
        case_type_cols=case_type_cols,
        case_type_col_order=case_type_col_order,
        genetic_distance_protocols=genetic_distance_protocols,
        tree_algorithms=tree_algorithms,
        case_type_access_abacs=case_type_access_abacs,
        case_type_share_abacs=case_type_share_abacs,
    )



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later


from typing import Any, ClassVar
from uuid import UUID

from pydantic import Field

from gen_epix.casedb.domain import enum
from gen_epix.commondb.domain.model.base import Model
from gen_epix.commondb.domain.model.organization import DataCollection, IdentifierIssuer
from gen_epix.fastapp.domain import Entity, create_links


class Subject(Model):
    """
    A person context bound to a particular data collection.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="subjects",
        table_name="subject",
        persistable=True,
        links=create_links(
            {
                1: ("data_collection_id", DataCollection, "data_collection"),
            }
        ),
    )
    data_collection_id: UUID = Field(
        description="The ID of the data collection. FOREIGN KEY"
    )
    data_collection: DataCollection | None = Field(
        default=None, description="The data collection"
    )
    external_ids: dict[IdentifierIssuer, str] | None = Field(
        default=None, description="A dictionary of external identifiers for the subject"
    )
    content: dict[str, Any] = Field(
        description="A dictionary containing the content of the subject"
    )


class SubjectIdentifier(Model):
    """
    A unique identifier for a subject, generated by a particular identifier
    issuer.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="subject_identifiers",
        table_name="subject_identifier",
        persistable=True,
        links=create_links(
            {
                1: ("subject_id", Subject, "subject"),
                2: ("identifier_issuer_id", IdentifierIssuer, "identifier_issuer"),
            }
        ),
    )
    subject_id: UUID = Field(description="The ID of the subject. FOREIGN KEY")
    subject: Subject | None = Field(default=None, description="The subject")
    identifier_issuer_id: UUID = Field(
        description="The ID of the identifier issuer. FOREIGN KEY"
    )
    identifier_issuer: IdentifierIssuer | None = Field(
        default=None, description="The identifier issuer"
    )
    identifier: str = Field(description="The identifier")



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later


from typing import Any, ClassVar, Self
from uuid import UUID

from pydantic import Field, model_validator

from gen_epix.casedb.domain import enum
from gen_epix.commondb.domain.model.base import Model
from gen_epix.fastapp.domain import Entity, create_keys, create_links


class ConceptSet(Model):
    """
    A set of concepts in the ontology.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="concept_sets",
        table_name="concept_set",
        persistable=True,
        keys=create_keys({1: "name"}),
    )
    code: str = Field(description="The code of the concept set", max_length=255)
    name: str = Field(description="The name of the concept set", max_length=255)
    type: enum.ConceptSetType = Field(description="The type of the concept set")
    regex: str | None = Field(
        default=None,
        description=(
            "The regular expression describing the concept set,"
            " in case of type REGULAR_EXPRESSION"
        ),
    )
    schema_definition: str | None = Field(
        default=None,
        description=(
            "The definition of the schema describing the concept set,"
            " in case of type CONTEXT_FREE_GRAMMAR_XXX"
        ),
    )
    schema_uri: str | None = Field(
        default=None,
        description=(
            "The URI to the schema describing the concept set,"
            " in case of type CONTEXT_FREE_GRAMMAR_XXX"
        ),
    )
    description: str | None = Field(
        default=None, description="The description of the concept set."
    )

    @model_validator(mode="after")
    def _validate_state(self) -> Self:
        if self.type == enum.ConceptSetType.REGULAR_LANGUAGE:
            if not self.regex:
                raise AssertionError(f"Type {self.type.value} requires regex")
        elif self.type in {
            enum.ConceptSetType.CONTEXT_FREE_GRAMMAR_JSON,
            enum.ConceptSetType.CONTEXT_FREE_GRAMMAR_XML,
        }:
            if not self.schema_definition and not self.schema_uri:
                raise AssertionError(
                    f"Type {self.type.value} requires schema_definition or schema_uri"
                )
        if self.schema_definition and self.schema_uri:
            raise AssertionError(
                "Only one of schema_definition or schema_uri can be set"
            )
        return self


class Concept(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="concepts",
        table_name="concept",
        persistable=True,
        keys=create_keys({1: ("concept_set_id", "code")}),
        links=create_links({1: ("concept_set_id", ConceptSet, "concept_set")}),
    )
    concept_set_id: UUID = Field(description="The ID of the concept set. FOREIGN KEY")
    concept_set: ConceptSet | None = Field(default=None, description="The concept set.")
    code: str = Field(description="Concept code within the set", max_length=255)
    name: str | None = Field(
        default=None, description="The name of the concept.", max_length=255
    )
    description: str | None = Field(
        default=None, description="The description of the concept."
    )
    rank: int | None = Field(
        default=None,
        description="The rank of the concept within the set. Must be provided for ordinal sets, for other sets it is optional and can be used for sorting.",
    )
    props: dict[str, Any] = Field(
        default_factory=dict, description="Additional properties of the concept."
    )


class ConceptRelation(Model):
    """
    A relation between two concepts (analogous to RegionRelation).
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="concept_relations",
        table_name="concept_relation",
        persistable=True,
        keys=create_keys({1: ("from_concept_id", "to_concept_id")}),
        links=create_links(
            {
                1: ("from_concept_id", Concept, "from_concept"),
                2: ("to_concept_id", Concept, "to_concept"),
            }
        ),
    )
    from_concept_id: UUID = Field(
        description="The ID of the first concept. FOREIGN KEY"
    )
    from_concept: Concept | None = Field(default=None, description="The first concept.")
    to_concept_id: UUID = Field(description="The ID of the second concept. FOREIGN KEY")
    to_concept: Concept | None = Field(default=None, description="The second concept.")
    relation: enum.ConceptRelationType = Field(
        description="The relation between the two concepts."
    )


class Disease(Model):
    """
    A disease.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="diseases",
        table_name="disease",
        persistable=True,
        keys=create_keys({1: "name"}),
    )
    name: str = Field(description="The name of the disease", max_length=255)
    icd_code: str | None = Field(
        default=None,
        description="The ICD code of the disease, if available",
        max_length=255,
    )


class EtiologicalAgent(Model):
    """
    An etiological agent.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="etiological_agents",
        table_name="etiological_agent",
        persistable=True,
        keys=create_keys({1: "name"}),
    )
    name: str = Field(description="The name of the etiological agent", max_length=255)
    type: str = Field(description="The type of the etiological agent", max_length=255)


class Etiology(Model):
    """
    The etiology of a disease based on an etiological agent.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="etiologies",
        table_name="etiology",
        persistable=True,
        keys=create_keys({1: ("disease_id", "etiological_agent_id")}),
        links=create_links(
            {
                1: ("disease_id", Disease, "disease"),
                2: ("etiological_agent_id", EtiologicalAgent, "etiological_agent"),
            }
        ),
    )
    disease_id: UUID = Field(description="The ID of the disease. FOREIGN KEY")
    disease: Disease | None = Field(default=None, description="The disease")
    etiological_agent_id: UUID = Field(
        description="The ID of the etiological agent. FOREIGN KEY"
    )
    etiological_agent: EtiologicalAgent | None = Field(
        None, description="The etiological agent"
    )



from collections.abc import Callable
from enum import Enum
from typing import Any

import sqlalchemy as sa
from pydantic import BaseModel
from pydantic.fields import FieldInfo
from sqlalchemy.orm import Mapped, MappedColumn, mapped_column

from gen_epix.commondb.domain.model import Model
from gen_epix.fastapp import Domain, Entity
from gen_epix.fastapp.repositories import create_sa_type_from_field_info
from gen_epix.fastapp.repositories.sa.util import get_sa_type_kwargs_from_field_info


def create_table_args(
    model_class: type[Model],
    field_name_map: dict[str, str] | None = None,
    **kwargs: Any,
) -> tuple:
    assert model_class.ENTITY is not None
    entity: Entity = model_class.ENTITY
    uq_constraints = []
    for field_names in entity.get_keys_field_names():
        sa_field_names = (
            [field_name_map.get(x, x) for x in field_names]
            if field_name_map
            else field_names
        )
        sa_field_name_str = "_".join(sa_field_names)
        uq_constraints.append(
            sa.UniqueConstraint(
                *sa_field_names,
                name=f"uq_{entity.table_name}_{sa_field_name_str}",
                **kwargs,
            )
        )
    if entity.schema_name:
        return entity.table_name, tuple(
            [*uq_constraints, {"schema": entity.schema_name}]
        )
    return entity.table_name, tuple([*uq_constraints])


def create_mapped_column(
    domain: Domain,
    model_class: type[Model],
    field_name: str,
    field_name_map: dict[type[Model], dict[str, str]] | None = None,
    **kwargs: Any,
) -> MappedColumn[Any]:
    assert model_class.ENTITY is not None
    entity: Entity = model_class.ENTITY
    field_info: FieldInfo = model_class.model_fields[field_name]
    sa_type = create_sa_type_from_field_info(field_info, field_info.annotation)
    nullable = kwargs.get("nullable", not field_info.is_required())
    doc = kwargs.pop("doc", field_info.description)
    link_entity = entity.get_link_entity(field_name)
    if link_entity and domain.get_service_type_for_entity(
        link_entity
    ) != domain.get_service_type_for_entity(entity):
        # Create foreign keys only within the same service
        link_entity = None
    ondelete = kwargs.pop("ondelete", None)
    onupdate = kwargs.pop("onupdate", None)
    if field_name_map is None:
        field_name_map = {}
    if model_class not in field_name_map:
        sa_field_name = field_name
    else:
        sa_field_name = field_name_map[model_class].get(field_name, field_name)
    fk_name = kwargs.pop("fk_name", f"fk_{entity.table_name}_{sa_field_name}")
    if link_entity:
        link_model_class = link_entity.model_class
        link_sa_id_field_name = (
            field_name_map[link_model_class][link_entity.id_field_name]
            if field_name_map
            else link_entity.id_field_name
        )
        ref_column_name = (
            f"{link_entity.schema_name}.{link_entity.table_name}.{link_sa_id_field_name}"
            if link_entity.schema_name
            else f"{link_entity.table_name}.{link_sa_id_field_name}"
        )
        return mapped_column(
            sa_type,
            sa.ForeignKey(
                ref_column_name, ondelete=ondelete, onupdate=onupdate, name=fk_name
            ),
            nullable=nullable,
            doc=doc,
            **kwargs,
        )
    return mapped_column(
        sa_type,
        nullable=nullable,
        primary_key=entity.id_field_name == field_name,
        doc=doc,
        **kwargs,
    )


def create_field_metadata(
    domain: Domain,
) -> tuple[
    dict[type[Model], list[str]],
    dict[type[Model], list[str]],
    dict[type[Model], Callable[[Any, Any], dict[str, Any]]],
]:
    model_classes: frozenset[type[Model]] = domain.models  # type: ignore[assignment]
    service_metadata_fields: dict[type[Model], list[str]] = {
        model_class: ["_modified_by"] for model_class in model_classes
    }
    db_metadata_fields: dict[type[Model], list[str]] = {
        model_class: [
            "_created_at",
            "_modified_at",
            "_version",
        ]
        for model_class in model_classes
    }
    generate_service_metadata: dict[
        type[Model], Callable[[Any, Any], dict[str, Any]]
    ] = {
        model_class: lambda x, y: {
            "_modified_by": y,
        }
        for model_class in model_classes
    }
    return service_metadata_fields, db_metadata_fields, generate_service_metadata


def set_entity_repository_model_classes(
    domain: Domain,
    sa_models_by_service_type: dict[Enum, dict[type[BaseModel], type]],
    row_metadata_mixin_class: type,
    field_name_map: dict[type, dict[str, str]] | None = None,
) -> None:
    if field_name_map is None:
        field_name_map = {}
    sa_metadata_field_names = set(row_metadata_mixin_class.__annotations__.keys()) - {
        "id"
    }
    sa_model_map = {}
    for curr_sa_model_map in sa_models_by_service_type.values():
        for model_class, sa_model_class in curr_sa_model_map.items():
            if model_class in sa_model_map:
                raise ValueError(f"Duplicate SA model for {model_class.__name__}")
            sa_model_map[model_class] = sa_model_class
    for entity in domain.get_dag_sorted_entities():
        if not entity.persistable:
            continue
        model_class = entity.model_class
        sa_model_class = sa_model_map.get(model_class)
        if not sa_model_class:
            raise ValueError(
                f"Model {model_class.__name__} does not have a corresponding SA model"
            )
        entity.set_db_model_class(sa_model_class)
        # Verify that the SA model has exactly the same fields as the model
        field_names = set(entity.get_field_names())
        relationship_field_names = set(entity.get_relationship_field_names())
        curr_field_name_map = field_name_map.get(model_class)
        if curr_field_name_map:
            field_names = {curr_field_name_map.get(x, x) for x in field_names}
        sa_field_names = (
            set(sa_model_class.__table__.columns.keys())
            - sa_metadata_field_names
            - relationship_field_names
        )
        extra_field_names = field_names - sa_field_names - relationship_field_names
        extra_field_names = {
            x for x in extra_field_names if f"{x}_id" not in field_names
        }
        if extra_field_names:
            extra_field_names_str = ", ".join(extra_field_names)
            raise ValueError(
                f"Model {model_class.__name__} has fields {extra_field_names_str} that are not in SA model {sa_model_class.__name__}"
            )
        extra_sa_field_names = sa_field_names - field_names
        if extra_sa_field_names:
            extra_field_names_str = ", ".join(extra_sa_field_names)
            raise ValueError(
                f"SA model {sa_model_class.__name__} has fields {extra_sa_field_names_str} that are not in model {model_class.__name__}"
            )


def get_mixin_mapped_column(
    model_mixin_class: type,
    field_name: str,
    sa_column_type: type[sa.types.TypeEngine],
    **kwargs: Any,
) -> Mapped:

    field_info: FieldInfo = getattr(model_mixin_class, field_name)
    annotation = model_mixin_class.__annotations__[field_name]
    # Extract SA arguments from mixin class based on sa_type
    kwargs["nullable"] = kwargs.get(  # pyright: ignore[reportArgumentType]
        "nullable", not field_info.is_required()
    )
    sa_column_type_kwargs = kwargs.pop(
        "sa_column_type_kwargs",
        get_sa_type_kwargs_from_field_info(sa_column_type, field_info),
    )
    # Create and return mapped column
    return mapped_column(
        create_sa_type_from_field_info(field_info, annotation, **sa_column_type_kwargs),
        **kwargs,
    )



import sqlalchemy.orm as orm
from sqlalchemy.orm import declarative_mixin

from gen_epix.commondb.domain import DOMAIN, enum, model
from gen_epix.commondb.repositories.sa_model.base import RowMetadataMixin
from gen_epix.commondb.repositories.sa_model.util import (
    create_mapped_column,
    create_table_args,
)

Base: type = orm.declarative_base(name=enum.ServiceType.ABAC.value)


@declarative_mixin
class OrganizationAdminPolicyMixin(RowMetadataMixin):
    organization_id = create_mapped_column(
        DOMAIN, model.OrganizationAdminPolicy, "organization_id"
    )
    user_id = create_mapped_column(DOMAIN, model.OrganizationAdminPolicy, "user_id")
    is_active = create_mapped_column(DOMAIN, model.OrganizationAdminPolicy, "is_active")


class OrganizationAdminPolicy(Base, OrganizationAdminPolicyMixin):
    __tablename__, __table_args__ = create_table_args(model.OrganizationAdminPolicy)



from gen_epix.commondb.domain.enum import Role as CommonRole
from gen_epix.commondb.domain.policy import RoleGenerator as CommonRoleGenerator
from gen_epix.fastapp.services.rbac import BaseRbacService
from gen_epix.omopdb.domain import command
from gen_epix.omopdb.domain.enum import Role


class RoleGenerator(CommonRoleGenerator):

    COMMON_ROLE_ENUM_MAP = {x: Role[x.name] for x in CommonRole}

    EXTRA_ROLE_SET_MAP = {}

    COMMON_ROLE_PERMISSION_SETS = (
        CommonRoleGenerator.map_from_common_role_permission_sets(
            COMMON_ROLE_ENUM_MAP, command.COMMON_COMMAND_MAP
        )
    )

    ROLE_PERMISSION_SETS = {
        # TODO: fill in permissions
        Role.APP_ADMIN: COMMON_ROLE_PERMISSION_SETS[Role.APP_ADMIN] | set(),
        Role.REFDATA_ADMIN: COMMON_ROLE_PERMISSION_SETS[Role.REFDATA_ADMIN] | set(),
        Role.ORG_ADMIN: COMMON_ROLE_PERMISSION_SETS[Role.ORG_ADMIN] | set(),
        Role.ORG_USER: COMMON_ROLE_PERMISSION_SETS[Role.ORG_USER] | set(),
        Role.GUEST: COMMON_ROLE_PERMISSION_SETS[Role.GUEST] | set(),
    }

    ROLE_HIERARCHY = CommonRoleGenerator.map_from_common_role_hierarchy(
        COMMON_ROLE_ENUM_MAP
    )

    ROLE_PERMISSIONS = BaseRbacService.expand_hierarchical_role_permissions(
        ROLE_HIERARCHY, ROLE_PERMISSION_SETS
    )



import importlib
import importlib.metadata
import re
import string
import tomllib
from typing import Any

from cachetools import TTLCache, cached

from gen_epix.commondb import policies
from gen_epix.commondb.app_impl_details import AppImplDetails
from gen_epix.commondb.domain import command, model
from gen_epix.commondb.domain.policy import BaseHasSystemOutagePolicy
from gen_epix.commondb.domain.service import BaseSystemService
from gen_epix.commondb.util import get_project_root
from gen_epix.fastapp import CrudOperation, EventTiming
from gen_epix.fastapp.app import App


class SystemService(BaseSystemService):
    REQUIREMENTS_FILE_NAME = "pyproject.toml"

    def __init__(self, app: App, **kwargs: Any) -> None:
        super().__init__(app, **kwargs)
        app_impl: AppImplDetails = app.impl
        self.has_system_outage_policy_class: type[BaseHasSystemOutagePolicy] = (
            app_impl.get_mapped_class(policies.HasSystemOutagePolicy)
        )

    def register_policies(self) -> None:
        """
        Registers policies that checks if the system has a current outage

        """
        policy = self.has_system_outage_policy_class(system_service=self)
        for command_class in self.app.domain.commands:
            self.app.register_policy(command_class, policy, EventTiming.BEFORE)

    def retrieve_outages(
        self, cmd: command.RetrieveOutagesCommand
    ) -> list[model.Outage]:
        with self.repository.uow() as uow:
            outages: list[model.Outage] = (
                self.repository.crud(  # type:ignore[assignment]
                    uow,
                    None,
                    model.Outage,
                    None,
                    None,
                    CrudOperation.READ_ALL,
                )
            )
        return outages

    def retrieve_licenses(
        self, cmd: command.RetrieveLicensesCommand
    ) -> list[model.PackageMetadata]:
        packages = SystemService._parse_and_get_package_metadata()
        return packages

    @staticmethod
    @cached(cache=TTLCache(maxsize=1000, ttl=60))
    def _parse_and_get_package_metadata() -> list[model.PackageMetadata]:
        """
        Parse pyproject.toml, extract package names, and get their metadata.
        """
        pyproject_path = get_project_root() / SystemService.REQUIREMENTS_FILE_NAME
        packages: list[model.PackageMetadata] = []

        if not pyproject_path.exists():
            return packages

        with open(pyproject_path, "rb") as f:
            pyproject_data = tomllib.load(f)

        packages.append(
            model.PackageMetadata(
                name="gen-epix-api",
                version=pyproject_data["project"].get("version", None),
                license=pyproject_data["project"].get("license", None),
                homepage=pyproject_data["project"]["urls"].get("Homepage", None),
            )
        )

        # Extract dependencies from project.dependencies
        dependencies = pyproject_data.get("project", {}).get("dependencies", [])

        for dependency in dependencies:
            # Extract package name (everything before version specifiers)
            match = re.match(r"^([a-zA-Z0-9_-]+)", dependency)
            if match:
                package_name = match.group(1)
                # Get metadata for this package
                try:
                    metadata = importlib.metadata.metadata(package_name)

                    # Try to get homepage from Home-page field first
                    homepage = metadata.get("Home-page", "")

                    # If no Home-page, extract from Project-URL using well-known labels
                    if not homepage:
                        project_urls = metadata.get("Project-URL", "")
                        homepage = SystemService._extract_homepage_from_project_urls(
                            project_urls
                        )

                    package_metadata = model.PackageMetadata(
                        name=metadata.get("Name", package_name),
                        version=metadata.get("Version", ""),
                        license=metadata.get("License"),
                        homepage=homepage or None,  # Convert empty string to None
                    )
                    packages.append(package_metadata)
                except importlib.metadata.PackageNotFoundError:
                    # Package not installed or name doesn't match, skip it
                    continue

        return packages

    @staticmethod
    def _normalize_project_url_label(label: str) -> str:
        """Normalize project URL label according to PEP 753."""
        chars_to_remove = string.punctuation + string.whitespace
        removal_map = str.maketrans("", "", chars_to_remove)
        return label.translate(removal_map).lower()

    @staticmethod
    def _extract_homepage_from_project_urls(project_urls_str: str) -> str:
        """
        Extract homepage URL from Project-URL metadata using well-known labels:
        (https://packaging.python.org/en/latest/specifications/well-known-project-urls/#well-known-labels)
        """
        if not project_urls_str:
            return ""

        # Parse Project-URL entries (format: "label, url")
        urls = {}
        for entry in project_urls_str.split("\n"):
            entry = entry.strip()
            if ", " in entry:
                label, url = entry.split(", ", 1)
                normalized_label = SystemService._normalize_project_url_label(label)
                urls[normalized_label] = url.strip()

        # Priority order based on well-known labels for homepage
        priority_labels = [
            "homepage",
            "documentation",
            "docs",
            "repository",
            "sourcecode",
            "github",
            "source",
        ]

        for label in priority_labels:
            if label in urls:
                return urls[label]

        # If no well-known labels found, return first URL available
        return next(iter(urls.values())) if urls else ""



from collections.abc import Iterable
from uuid import UUID

import gen_epix.casedb.domain.command as command
import gen_epix.casedb.domain.enum as enum
import gen_epix.casedb.domain.model as model
from gen_epix.casedb.domain import exc
from gen_epix.casedb.domain.policy import BaseCaseAbacPolicy
from gen_epix.casedb.domain.service import BaseCaseService as DomainBaseCaseService
from gen_epix.casedb.services.case.base import BaseCaseService
from gen_epix.casedb.services.case.case_transformer import CaseTransformer
from gen_epix.casedb.services.case.create_case import (
    case_service_create_case_set,
    case_service_create_cases,
)
from gen_epix.casedb.services.case.create_seq import (
    case_service_create_file_for_read_set_or_seq,
    case_service_create_read_sets_or_seqs_for_cases,
)
from gen_epix.casedb.services.case.crud import crud
from gen_epix.casedb.services.case.read_association_with_valid_ids import (
    case_service_read_association_with_valid_ids,
)
from gen_epix.casedb.services.case.retrieve_case import (
    case_service_retrieve_cases_by_id,
    case_service_retrieve_cases_by_query,
)
from gen_epix.casedb.services.case.retrieve_complete_case_type import (
    case_service_retrieve_complete_case_type,
)
from gen_epix.casedb.services.case.retrieve_seq import (
    case_service_retrieve_assembly_protocols,
    case_service_retrieve_genetic_sequence_by_case,
    case_service_retrieve_genetic_sequence_fasta_by_case,
    case_service_retrieve_library_prep_protocols,
    case_service_retrieve_phylogenetic_tree,
)
from gen_epix.casedb.services.case.retrieve_stats import (
    case_service_retrieve_case_set_stats,
    case_service_retrieve_case_type_stats,
)
from gen_epix.commondb.util import map_paired_elements
from gen_epix.fastapp import BaseUnitOfWork, CrudOperation
from gen_epix.filter import Filter, UuidSetFilter
from gen_epix.filter.composite import CompositeFilter
from gen_epix.filter.datetime_range import DatetimeRangeFilter
from gen_epix.filter.enum import LogicalOperator


class CaseService(BaseCaseService):

    def crud(  # type:ignore[override]
        self, cmd: command.CrudCommand
    ) -> list[model.Model] | model.Model | list[UUID] | UUID | list[bool] | bool | None:
        return crud(self, cmd)

    def validate_cases(
        self, cmd: command.ValidateCasesCommand
    ) -> model.CaseValidationReport:
        case_type_id = cmd.case_type_id
        created_in_data_collection_id = cmd.created_in_data_collection_id

        # @ABAC: verify if case set or cases may be created in the given data collection(s)
        case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)
        assert case_abac is not None
        is_allowed = case_abac.is_allowed(
            case_type_id,
            enum.CaseRight.ADD_CASE,
            True,
            created_in_data_collection_id=created_in_data_collection_id,
            tgt_data_collection_ids=cmd.data_collection_ids,
        )
        if not is_allowed:
            assert cmd.user is not None
            raise exc.UnauthorizedAuthError(
                f"User {cmd.user.id} is not allowed to create a case set/cases in the given data collection(s)"
            )

        # TODO: merge data with existing cases when updating
        curr_cmd = command.RetrieveCompleteCaseTypeCommand(
            user=cmd.user, case_type_id=case_type_id
        )
        curr_cmd._policies.extend(cmd._policies)
        complete_case_type = case_service_retrieve_complete_case_type(self, curr_cmd)
        transformer = CaseTransformer(self, complete_case_type)
        transform_result = transformer(cmd)
        if not transform_result.success:
            raise exc.DataException(f"Failed to transform case data")
        case_validation_report: model.CaseValidationReport = (
            transform_result.transformed_object  # type:ignore[assignment]
        )

        return case_validation_report

    def create_cases(self, cmd: command.CreateCasesCommand) -> list[model.Case] | None:
        return case_service_create_cases(self, cmd)

    def create_case_set(
        self, cmd: command.CreateCaseSetCommand
    ) -> model.CaseSet | None:
        return case_service_create_case_set(self, cmd)

    def create_reads_sets_for_cases(
        self, cmd: command.CreateReadSetsForCasesCommand
    ) -> list[model.ReadSet]:
        retval: list[model.ReadSet] = case_service_create_read_sets_or_seqs_for_cases(
            self, cmd
        )  # type:ignore[assignment]
        return retval

    def create_seqs_for_cases(
        self, cmd: command.CreateSeqsForCasesCommand
    ) -> list[model.Seq]:
        retval: list[model.Seq] = case_service_create_read_sets_or_seqs_for_cases(
            self, cmd
        )  # type:ignore[assignment]
        return retval

    def create_file_for_read_set(
        self, cmd: command.CreateFileForReadSetCommand
    ) -> UUID:
        return case_service_create_file_for_read_set_or_seq(self, cmd)

    def create_file_for_seq(self, cmd: command.CreateFileForSeqCommand) -> UUID:
        return case_service_create_file_for_read_set_or_seq(self, cmd)

    def retrieve_complete_case_type(
        self: DomainBaseCaseService,
        cmd: command.RetrieveCompleteCaseTypeCommand,
    ) -> model.CompleteCaseType:
        return case_service_retrieve_complete_case_type(self, cmd)

    def retrieve_case_type_stats(
        self,
        cmd: command.RetrieveCaseTypeStatsCommand,
    ) -> list[model.CaseTypeStat]:
        return case_service_retrieve_case_type_stats(self, cmd)

    def retrieve_case_set_stats(
        self,
        cmd: command.RetrieveCaseSetStatsCommand,
    ) -> list[model.CaseSetStat]:
        return case_service_retrieve_case_set_stats(self, cmd)

    def retrieve_cases_by_query(
        self, cmd: command.RetrieveCasesByQueryCommand
    ) -> list[UUID]:
        return case_service_retrieve_cases_by_query(self, cmd)

    def retrieve_cases_by_id(
        self, cmd: command.RetrieveCasesByIdCommand
    ) -> list[model.Case]:
        return case_service_retrieve_cases_by_id(self, cmd)

    def retrieve_case_or_set_rights(
        self,
        cmd: command.RetrieveCaseRightsCommand | command.RetrieveCaseSetRightsCommand,
    ) -> list[model.CaseRights] | list[model.CaseSetRights]:
        is_case_set = isinstance(cmd, command.RetrieveCaseSetRightsCommand)
        case_or_set_ids = cmd.case_set_ids if is_case_set else cmd.case_ids  # type: ignore[union-attr]
        user, repository = self._get_user_and_repository(cmd)

        # Special case: zero case_ids
        if not case_or_set_ids:
            return []

        # @ABAC: get case abac
        case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)
        assert case_abac is not None

        # Retrieve all cases and case data collection links
        with repository.uow() as uow:
            # Retrieve cases/sets
            cases_or_sets: list[model.CaseSet] | list[model.Case] = self.repository.crud(  # type: ignore[assignment]
                uow,
                user.id,
                model.CaseSet if is_case_set else model.Case,
                None,
                case_or_set_ids,
                CrudOperation.READ_SOME,
            )
            # Retrieve case/set data collection links
            key = "case_set_id" if is_case_set else "case_id"
            case_or_set_data_collection_links: list[model.CaseDataCollectionLink] | list[model.CaseSetDataCollectionLink] = self.repository.crud(  # type: ignore[assignment]
                uow,
                user.id,
                (
                    model.CaseSetDataCollectionLink
                    if is_case_set
                    else model.CaseDataCollectionLink
                ),
                None,
                None,
                CrudOperation.READ_ALL,
                filter=UuidSetFilter(
                    key=key,
                    members=frozenset(case_or_set_ids),
                ),
            )

        # Determine case/set rights
        case_or_set_data_collections: dict[UUID, set[UUID]] = map_paired_elements(  # type: ignore[assignment]
            (
                (x.case_set_id if is_case_set else x.case_id, x.data_collection_id)  # type: ignore[union-attr]
                for x in case_or_set_data_collection_links
            ),
            as_set=True,
        )

        # Generate return value
        retval: list[model.CaseSetRights] | list[model.CaseRights] = []
        for case_or_set in cases_or_sets:
            assert case_or_set.id is not None
            data_collection_ids = case_or_set_data_collections.get(
                case_or_set.id, set()
            )
            data_collection_ids.add(case_or_set.created_in_data_collection_id)
            args: tuple = (
                case_or_set.id,
                case_or_set.case_type_id,
                case_or_set.created_in_data_collection_id,
                case_or_set_data_collections.get(case_or_set.id, set()),
            )
            retval.append(case_abac.get_case_set_rights(*args) if is_case_set else case_abac.get_case_rights(*args))  # type: ignore[arg-type]

        return retval

    def retrieve_phylogenetic_tree(
        self, cmd: command.RetrievePhylogeneticTreeByCasesCommand
    ) -> model.PhylogeneticTree:
        return case_service_retrieve_phylogenetic_tree(self, cmd)

    def retrieve_genetic_sequence_by_case(
        self,
        cmd: command.RetrieveGeneticSequenceByCaseCommand,
    ) -> list[model.GeneticSequence]:
        return case_service_retrieve_genetic_sequence_by_case(self, cmd)

    def retrieve_genetic_sequence_fasta_by_case(
        self, cmd: command.RetrieveGeneticSequenceFastaByCaseCommand
    ) -> Iterable[str]:
        """
        Return a streaming iterable of FASTA formatted lines.
        Path:
        HTTP client
        -> casedb endpoint
        -> casedb service calls casedb seqdb command
        -> seqdb command (inside casedb) calls ext_app with RetrieveSeqFastaCommand
        -> seqdb service calls correct repository (dict or SA implementation) to stream Seq rows
        -> seqdb service converts rows to FASTA lines on the fly
        -> returns an iterator
        -> casedb forwards that iterator
        -> FastAPI wraps it in a StreamingResponse.
        """
        return case_service_retrieve_genetic_sequence_fasta_by_case(self, cmd)

    def retrieve_library_prep_protocols(
        self,
        cmd: command.RetrieveLibraryPrepProtocolsCommand,
    ) -> list[model.LibraryPrepProtocol]:
        return case_service_retrieve_library_prep_protocols(self, cmd)

    def retrieve_assembly_protocols(
        self, cmd: command.RetrieveAssemblyProtocolsCommand
    ) -> list[model.AssemblyProtocol]:
        return case_service_retrieve_assembly_protocols(self, cmd)

    def _read_association_with_valid_ids(
        self,
        command_class: type[command.CrudCommand],
        field_name1: str,
        field_name2: str,
        valid_ids1: set[UUID] | frozenset[UUID] | None = None,
        valid_ids2: set[UUID] | frozenset[UUID] | None = None,
        match_all1: bool = False,
        match_all2: bool = False,
        return_type: str = "objects",
        uow: BaseUnitOfWork | None = None,
        user: model.User | None = None,
    ) -> list[model.Model] | list[UUID] | dict[UUID, set[UUID]]:
        return case_service_read_association_with_valid_ids(
            self,
            command_class,
            field_name1,
            field_name2,
            valid_ids1=valid_ids1,
            valid_ids2=valid_ids2,
            match_all1=match_all1,
            match_all2=match_all2,
            return_type=return_type,
            uow=uow,
            user=user,
        )

    def _retrieve_case_sets_with_content_right(
        self,
        uow: BaseUnitOfWork,
        user_id: UUID,
        case_abac: model.CaseAbac,
        right: enum.CaseRight,
        case_set_ids: list[UUID] | None = None,
        case_type_ids: set[UUID] | None = None,
        filter: Filter | None = None,
        on_invalid_case_set_id: str = "raise",
    ) -> list[model.CaseSet]:
        # TODO: This is a temporary implementation, to be replaced by optimized query
        if right not in enum.CaseRightSet.CASE_SET_CONTENT.value:
            raise exc.InvalidArgumentsError(f"Invalid case abac right: {right.value}")
        if on_invalid_case_set_id not in {"raise", "ignore"}:
            raise exc.InvalidArgumentsError(
                f"Invalid on_invalid_case_set_id: {on_invalid_case_set_id}"
            )

        # Retrieve all case sets, potentially filtered
        case_sets: list[model.CaseSet]
        if case_set_ids:
            if filter:
                raise exc.InvalidArgumentsError(
                    "Cannot use datetime range filter with case set ids"
                )
            case_sets = self.repository.crud(  # type:ignore[assignment]
                uow,
                user_id,
                model.CaseSet,
                None,
                case_set_ids,
                CrudOperation.READ_SOME,
            )
        else:
            case_sets = self.repository.crud(  # type:ignore[assignment]
                uow,
                user_id,
                model.CaseSet,
                None,
                None,
                CrudOperation.READ_ALL,
                filter=filter,
            )

        # Filter on case_type_ids if any or verify that all case sets have a valid
        # case_type_id if case_set_ids is given
        # TODO: add more efficient implementation by adding this as a filter in the
        # call to the repository
        if case_type_ids is not None:
            if case_set_ids:
                if on_invalid_case_set_id == "raise":
                    if not all(x.case_type_id in case_type_ids for x in case_sets):
                        raise exc.InvalidArgumentsError(
                            f"Some case sets have invalid case type ids: {case_set_ids}"
                        )
                elif on_invalid_case_set_id == "ignore":
                    pass
                else:
                    raise AssertionError(
                        f"Invalid on_invalid_case_set_id: {on_invalid_case_set_id}"
                    )
            case_sets = [x for x in case_sets if x.case_type_id in case_type_ids]

        # Special case: full_access
        if case_abac.is_full_access:
            return case_sets

        # @ABAC: filter case sets to which the user has read access
        case_set_data_collections = self._retrieve_case_set_data_collections_map(
            uow, user_id
        )
        has_access = case_abac.get_combinations_with_access_right(right)
        filtered_case_sets = []
        for case_set in case_sets:
            # Check if user has any access to case
            case_type_id = case_set.case_type_id
            if case_type_id not in has_access:
                if case_set_ids:
                    if on_invalid_case_set_id == "raise":
                        raise exc.UnauthorizedAuthError(
                            f"User {user_id} has no access to some requested cases"
                        )
                    elif on_invalid_case_set_id == "ignore":
                        pass
                    else:
                        raise AssertionError(
                            f"Invalid on_invalid_case_id: {on_invalid_case_set_id}"
                        )
                continue
            # Check if user has access to any of the data collections of the case set
            data_collection_ids = case_set_data_collections.get(
                case_set.id, set()  # type:ignore[arg-type]
            )
            data_collection_ids.add(case_set.created_in_data_collection_id)
            if not data_collection_ids.intersection(has_access[case_type_id]):
                if case_set_ids:
                    if on_invalid_case_set_id == "raise":
                        raise exc.UnauthorizedAuthError(
                            f"User {user_id} has no access to some requested case sets"
                        )
                    elif on_invalid_case_set_id == "ignore":
                        pass
                    else:
                        raise AssertionError(
                            f"Invalid on_invalid_case_set_id: {on_invalid_case_set_id}"
                        )
                continue
            # Keep case
            filtered_case_sets.append(case_set)
        return filtered_case_sets

    def _retrieve_cases_with_content_right(
        self,
        uow: BaseUnitOfWork,
        user_id: UUID,
        case_abac: model.CaseAbac,
        right: enum.CaseRight,
        case_ids: list[UUID] | None = None,
        case_type_ids: set[UUID] | None = None,
        datetime_range_filter: DatetimeRangeFilter | None = None,
        on_invalid_case_id: str = "raise",
        filter_content: bool = True,
        extra_access_case_type_col_ids: set[UUID] | None = None,
    ) -> list[model.Case]:
        # TODO: This is a temporary implementation, to be replaced by optimized query
        if right not in enum.CaseRightSet.CASE_CONTENT.value:
            raise exc.InvalidArgumentsError(f"Invalid case abac right: {right.value}")
        if on_invalid_case_id not in {"raise", "ignore"}:
            raise exc.InvalidArgumentsError(
                f"Invalid on_invalid_case_id: {on_invalid_case_id}"
            )

        # Retrieve all cases, potentially filtered by datetime range
        if datetime_range_filter:
            if datetime_range_filter.key and datetime_range_filter.key != "case_date":
                raise exc.InvalidArgumentsError(
                    f"Invalid datetime range filter key: {datetime_range_filter.key}"
                )
            datetime_range_filter.key = "case_date"
        cases: list[model.Case]
        if case_ids:
            if datetime_range_filter:
                raise exc.InvalidArgumentsError(
                    "Cannot use datetime range filter with case ids"
                )
            cases = self.repository.crud(  # type:ignore[assignment]
                uow,
                user_id,
                model.Case,
                None,
                case_ids,
                CrudOperation.READ_SOME,
            )
        else:
            cases = self.repository.crud(  # type:ignore[assignment]
                uow,
                user_id,
                model.Case,
                None,
                None,
                CrudOperation.READ_ALL,
                filter=datetime_range_filter,
            )

        # Filter on case_type_ids if any or verify that all cases have a valid
        # case_type_id if case_ids is given
        # TODO: add more efficient implementation by adding this as a filter in the
        # call to the repository
        if case_type_ids is not None:
            if case_ids:
                if not all(x.case_type_id in case_type_ids for x in cases):
                    raise exc.InvalidArgumentsError(
                        f"Some cases have invalid case type ids: {case_ids}"
                    )
                if on_invalid_case_id == "raise":
                    if not all(x.case_type_id in case_type_ids for x in cases):
                        raise exc.InvalidArgumentsError(
                            f"Some cases have invalid case type ids: {case_ids}"
                        )
                elif on_invalid_case_id == "ignore":
                    pass
                else:
                    raise AssertionError(
                        f"Invalid on_invalid_case_id: {on_invalid_case_id}"
                    )
            cases = [x for x in cases if x.case_type_id in case_type_ids]

        # Special case: full_access
        if case_abac.is_full_access:
            return cases

        # @ABAC: filter cases to which the user has read access, and optionally also
        # the content (case type cols)
        case_data_collections = self._retrieve_case_data_collections_map(uow, user_id)
        has_access = case_abac.get_combinations_with_access_right(right)
        filtered_cases = []
        for case in cases:
            # Check if user has any access to case
            case_type_id = case.case_type_id
            if case_type_id not in has_access:
                if case_ids:
                    if on_invalid_case_id == "raise":
                        raise exc.UnauthorizedAuthError(
                            f"User {user_id} has no access to some requested cases"
                        )
                    elif on_invalid_case_id == "ignore":
                        pass
                    else:
                        raise AssertionError(
                            f"Invalid on_invalid_case_id: {on_invalid_case_id}"
                        )
                continue
            # Check if user has access to any data collection of the case
            data_collection_ids = case_data_collections.get(
                case.id, set()  # type:ignore[arg-type]
            )
            data_collection_ids.add(case.created_in_data_collection_id)
            if not data_collection_ids.intersection(has_access[case_type_id]):
                if case_ids:
                    if on_invalid_case_id == "raise":
                        raise exc.UnauthorizedAuthError(
                            f"User {user_id} has no access to some requested cases"
                        )
                    elif on_invalid_case_id == "ignore":
                        pass
                    else:
                        raise AssertionError(
                            f"Invalid on_invalid_case_id: {on_invalid_case_id}"
                        )
                continue
            # Keep case
            filtered_cases.append(case)
            # Continue to next case if case content need not be filtered
            if not filter_content:
                continue
            # Determine which case type cols the user has access to
            data_collection_col_access = case_abac.case_type_access_abacs[case_type_id]
            case_type_col_ids = set()
            for data_collection_id in data_collection_ids:
                # Add case type cols with access to the case for this data
                # collection
                case_type_access_abac = data_collection_col_access.get(
                    data_collection_id
                )
                if case_type_access_abac is not None:
                    case_type_col_ids.update(
                        case_type_access_abac.read_case_type_col_ids
                    )
            if extra_access_case_type_col_ids is not None:
                case_type_col_ids.update(extra_access_case_type_col_ids)
            if not case_type_col_ids:
                data_collection_ids_str = ", ".join(
                    [str(x) for x in data_collection_ids]
                )
                raise AssertionError(
                    f"User {user_id} has zero columns with {right.value} access to case {case.id}, data collections ({data_collection_ids_str}) even though the case has some {right.value} access"
                )
            # Filter case content
            case.content = {
                x: y for x, y in case.content.items() if x in case_type_col_ids
            }
        return filtered_cases

    def _retrieve_case_data_collections_map(
        self,
        uow: BaseUnitOfWork,
        user_id: UUID,
        case_ids: Iterable[UUID] | None = None,
        data_collection_ids: Iterable[UUID] | None = None,
    ) -> dict[UUID, set[UUID]]:
        return self._retrieve_association_map(
            uow,
            user_id,
            model.CaseDataCollectionLink,
            "case_id",
            "data_collection_id",
            obj_ids1=frozenset(case_ids) if case_ids else None,
            obj_ids2=frozenset(data_collection_ids) if data_collection_ids else None,
        )

    def _retrieve_case_set_data_collections_map(
        self,
        uow: BaseUnitOfWork,
        user_id: UUID,
        case_set_ids: Iterable[UUID] | None = None,
        data_collection_ids: Iterable[UUID] | None = None,
    ) -> dict[UUID, set[UUID]]:
        return self._retrieve_association_map(
            uow,
            user_id,
            model.CaseSetDataCollectionLink,
            "case_set_id",
            "data_collection_id",
            obj_ids1=frozenset(case_set_ids) if case_set_ids else None,
            obj_ids2=frozenset(data_collection_ids) if data_collection_ids else None,
        )

    def _retrieve_case_case_sets_map(
        self,
        uow: BaseUnitOfWork,
        user_id: UUID,
        case_ids: Iterable[UUID] | None = None,
        case_set_ids: Iterable[UUID] | None = None,
    ) -> dict[UUID, set[UUID]]:
        return self._retrieve_association_map(
            uow,
            user_id,
            model.CaseSetMember,
            "case_id",
            "case_set_id",
            obj_ids1=frozenset(case_ids) if case_ids else None,
            obj_ids2=frozenset(case_set_ids) if case_set_ids else None,
        )

    def _retrieve_association_map(
        self,
        uow: BaseUnitOfWork,
        user_id: UUID | None,
        association_class: type[model.Model],
        link_field_name1: str,
        link_field_name2: str,
        obj_ids1: frozenset[UUID] | None = None,
        obj_ids2: frozenset[UUID] | None = None,
    ) -> dict[UUID, set[UUID]]:
        """
        Get a dict[obj_id1, set[obj_ids]] based on the association stored in the association_class objs.
        """
        # Create a filter to restrict the association objs if necessary
        filter: Filter | None
        if obj_ids1:
            filter1 = UuidSetFilter(key=link_field_name1, members=obj_ids1)
        else:
            filter1 = None
        if obj_ids2:
            filter2 = UuidSetFilter(key=link_field_name2, members=obj_ids2)
        else:
            filter2 = None
        if filter1 and filter2:
            filter = CompositeFilter(
                filters=[filter1, filter2], operator=LogicalOperator.AND
            )
        elif filter1:
            filter = filter1
        elif filter2:
            filter = filter2
        else:
            filter = None
        # Retrieve association objs and convert to map
        value_pairs_iterable: Iterable[tuple[UUID, UUID]] = self.repository.read_fields(
            uow=uow,
            user_id=user_id,
            model_class=association_class,
            field_names=[link_field_name1, link_field_name2],
            filter=filter,
        )
        association_map: dict[UUID, set[UUID]] = map_paired_elements(value_pairs_iterable, as_set=True)  # type: ignore[assignment]

        return association_map

    def _retrieve_sequence_column_data(
        self, uow: BaseUnitOfWork, user: model.User, seq_case_type_col_id: UUID
    ) -> tuple[model.CaseTypeCol, model.Col]:
        repository = self.repository
        seq_case_type_col: model.CaseTypeCol = repository.crud(  # type: ignore[assignment]
            uow,
            user.id,
            model.CaseTypeCol,
            None,
            seq_case_type_col_id,
            CrudOperation.READ_ONE,
        )
        seq_col: model.Col = repository.crud(  # type: ignore[assignment]
            uow,
            user.id,
            model.Col,
            None,
            seq_case_type_col.col_id,
            CrudOperation.READ_ONE,
        )
        if seq_col.col_type != enum.ColType.GENETIC_SEQUENCE:
            raise exc.InvalidArgumentsError(
                f"Case type column {seq_col.id} is not of type {enum.ColType.GENETIC_SEQUENCE.value}"
            )
        return seq_case_type_col, seq_col

    def _verify_case_set_member_case_type(
        self, user: model.User, case_set_members: list[model.CaseSetMember]
    ) -> None:
        with self.repository.uow() as uow:
            case_set_ids = {x.case_set_id for x in case_set_members}
            case_ids = {x.case_id for x in case_set_members}
            case_sets_: list[model.CaseSet] = (
                self.repository.crud(  # type:ignore[assignment]
                    uow,
                    user.id if user else None,
                    model.CaseSet,
                    None,
                    list(case_set_ids),
                    CrudOperation.READ_SOME,
                )
            )
            case_sets = {x.id: x for x in case_sets_}
            cases_: list[model.Case] = self.repository.crud(  # type:ignore[assignment]
                uow,
                user.id if user else None,
                model.Case,
                None,
                list(case_ids),
                CrudOperation.READ_SOME,
            )
            cases = {x.id: x for x in cases_}
        invalid_case_set_member_ids = [
            x.id
            for x in case_set_members
            if case_sets[x.case_set_id].case_type_id != cases[x.case_id].case_type_id
        ]
        if invalid_case_set_member_ids:
            invalid_case_set_member_ids_str = ", ".join(
                [str(x) for x in invalid_case_set_member_ids]
            )
            raise exc.InvalidArgumentsError(
                f"Case set members invalid, case set and case must have the same case type: {invalid_case_set_member_ids_str}"
            )

    @staticmethod
    def _compose_id_filter(*key_and_ids: tuple[str, set[UUID]]) -> Filter:
        if len(key_and_ids) == 1:
            key, ids = key_and_ids[0]
            return UuidSetFilter(key=key, members=ids)  # type: ignore[arg-type]
        return CompositeFilter(
            filters=[
                UuidSetFilter(key=key, members=ids)  # type: ignore[arg-type]
                for key, ids in key_and_ids
            ],
            operator=LogicalOperator.AND,
        )



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later


from typing import Any
from uuid import UUID

import sqlalchemy.orm as orm
from sqlalchemy.orm import Mapped, relationship

from gen_epix.casedb.domain import DOMAIN, enum, model
from gen_epix.commondb.repositories.sa_model import (
    RowMetadataMixin,
    create_mapped_column,
    create_table_args,
)

Base: type = orm.declarative_base(name=enum.ServiceType.ONTOLOGY.value)


class ConceptSet(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.ConceptSet)

    code: Mapped[str] = create_mapped_column(DOMAIN, model.ConceptSet, "code")
    name: Mapped[str] = create_mapped_column(DOMAIN, model.ConceptSet, "name")
    type: Mapped[enum.ConceptSetType] = create_mapped_column(
        DOMAIN, model.ConceptSet, "type"
    )
    regex: Mapped[str] = create_mapped_column(DOMAIN, model.ConceptSet, "regex")
    schema_definition: Mapped[str] = create_mapped_column(
        DOMAIN, model.ConceptSet, "schema_definition"
    )
    schema_uri: Mapped[str] = create_mapped_column(
        DOMAIN, model.ConceptSet, "schema_uri"
    )
    description: Mapped[str] = create_mapped_column(
        DOMAIN, model.ConceptSet, "description"
    )


class Concept(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Concept)

    concept_set_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Concept, "concept_set_id"
    )

    code: Mapped[str] = create_mapped_column(DOMAIN, model.Concept, "code")
    name: Mapped[str] = create_mapped_column(DOMAIN, model.Concept, "name")
    description: Mapped[str] = create_mapped_column(
        DOMAIN, model.Concept, "description"
    )
    rank: Mapped[int] = create_mapped_column(DOMAIN, model.Concept, "rank")
    props: Mapped[dict[str, Any]] = create_mapped_column(DOMAIN, model.Concept, "props")


class ConceptRelation(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.ConceptRelation)

    from_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConceptRelation, "from_concept_id"
    )
    to_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConceptRelation, "to_concept_id"
    )
    relation: Mapped[enum.ConceptRelationType] = create_mapped_column(
        DOMAIN, model.ConceptRelation, "relation"
    )


class Disease(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Disease)

    name: Mapped[str] = create_mapped_column(DOMAIN, model.Disease, "name")
    icd_code: Mapped[str] = create_mapped_column(DOMAIN, model.Disease, "icd_code")


class EtiologicalAgent(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.EtiologicalAgent)

    name: Mapped[str] = create_mapped_column(DOMAIN, model.EtiologicalAgent, "name")
    type: Mapped[str] = create_mapped_column(DOMAIN, model.EtiologicalAgent, "type")


class Etiology(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Etiology)

    disease_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Etiology, "disease_id"
    )
    etiological_agent_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Etiology, "etiological_agent_id"
    )

    disease: Mapped[Disease] = relationship(Disease, foreign_keys=[disease_id])
    etiological_agent: Mapped[EtiologicalAgent] = relationship(
        EtiologicalAgent, foreign_keys=[etiological_agent_id]
    )
